#include <vector>
#include <assert.h>

#include "opengl.h"
#include "game.h"
#include "input.h"
#include "datapath.h"
#include "shader.h"
#include "sdrman.h"
#include "unistate.h"
#include "scr_debug.h"
#include "scr_game.h"
#include "logger.h"
#include "mesh.h"
#include "audio/audio.h"
#include "timer.h"
#include "options.h"

static void toggle_debug();

static int width, height;

static std::vector<BaseScreen*> screens;
static BaseScreen *debug_screen;

static void standard_state_uniforms();
static bool create_screens();

ShaderProg *defsdr;
Timer timer;

bool game_init()
{
	init_opengl();
	CHECKGLERR;

#ifdef __GLEW_H__
	/*
	if(GLEW_ARB_multisample) {
		info_log("enabling multisampling\n");
		glEnable(GL_MULTISAMPLE);
	} else {
		info_log("no multisampling support, will try old-style smoothing\n");
		// line smoothing enabled when needed (mostly by tools)
	}
	*/
	/*
#ifdef __unix__
	if(GLXEW_EXT_swap_control) {
		int val = 1;
		if(GLXEW_EXT_swap_control_tear) {
			val = -1;
		}
		debug_log("using glXSwapIntervalEXT to enable vsync\n");
		glXSwapIntervalEXT(glXGetCurrentDisplay(), glXGetCurrentDrawable(), val);
	} else if(GLXEW_SGI_swap_control) {
		debug_log("using glXSwapIntervalSGI to enable vsync\n");
		glXSwapIntervalSGI(1);
	}
#endif
#ifdef WIN32
	if(WGLEW_EXT_swap_control) {
		int val = 1;
		if(WGLEW_EXT_swap_control_tear) {
			val = -1;
		}
		debug_log("using wglSwapIntervalEXT to enable vsync\n");
		wglSwapIntervalEXT(val);
	}
#endif
*/
#endif	// __GLEW_H__

	// The configuration file is expected to be on the root
	// of our data archive.
	add_data_path("data");
	add_data_path("sdr");

	standard_state_uniforms();

	if(!(defsdr = get_sdrprog("default.v.glsl", "default.p.glsl"))) {
		error_log("failed to load default shader, aborting\n");
		return false;
	}

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);

	init_audio();

	if(!create_screens()) {
		error_log("failed to create scenes\n");
		return false;
	}
#if (defined(TARGET_IPHONE) || defined(ANDROID)) && !defined(NDEBUG)
	toggle_debug();
#endif

	return true;
}

void game_cleanup()
{
	destroy_audio();

	delete defsdr;

	// pop all screens
	while(current_screen()) {
		pop_screen();
	}
	// and then destroy them
	for(size_t i=0; i<screens.size(); i++) {
		delete screens[i];
	}
	delete debug_screen;
}

void game_display()
{
	unsigned long msec = timer.get_msec();

	// set current global time state uniform
	static int st_time_idx = -1;
	if(st_time_idx == -1) {
		st_time_idx = get_unistate_index("st_time");
	}
	set_unistate(st_time_idx, (float)msec / 1000.0f);

	current_screen()->update(msec);

	current_screen()->pre_draw();
	current_screen()->display();
	current_screen()->post_draw();

	swap_buffers();
	CHECKGLERR;
}

void game_reshape(int x, int y)
{
	width = x;
	height = y;

	// set framebuffer size uniform
	static int st_fbsize_idx = -1;
	if(st_fbsize_idx == -1) {
		st_fbsize_idx = get_unistate_index("st_fbsize");
	}
	set_unistate(st_fbsize_idx, Vector2(x, y));

	glViewport(0, 0, x, y);
	current_screen()->reshape(x, y);
}

void game_key(int key, bool pressed)
{
	// special key processing
	if(pressed) {
		switch(key) {
		case 27:
			exit(0);

		case '`':
		case '\\':
			toggle_debug();
			break;

		default:
			break;
		}
	}
	set_key_state(key, pressed);
	current_screen()->keyboard(key, pressed);
}

void game_mbutton(int id, int bn, bool state, int x, int y)
{
	current_screen()->button(bn, state, x, y);
	set_mouse_state(x, y, bn, state);	// this must be after the event dispatch
}

void game_mmotion(int id, int x, int y, bool pressed)
{
	set_mouse_state(x, y);
	current_screen()->motion(x, y, pressed);
}


int get_screen_width()
{
	return width;
}

int get_screen_height()
{
	return height;
}



static void standard_state_uniforms()
{
	// setup all the standard state uniforms
	add_unistate("st_world_matrix", ST_MATRIX4);
	add_unistate("st_world_matrix_transpose", ST_MATRIX4);
	//add_unistate("st_world_matrix_inverse", ST_MATRIX4);
	add_unistate("st_world_matrix3", ST_MATRIX3);

	add_unistate("st_view_matrix", ST_MATRIX4);
	add_unistate("st_view_matrix_transpose", ST_MATRIX4);
	//add_unistate("st_view_matrix_inverse", ST_MATRIX4);
	add_unistate("st_view_matrix3", ST_MATRIX3);

	add_unistate("st_proj_matrix", ST_MATRIX4);
	//add_unistate("st_proj_matrix_inverse", ST_MATRIX4);

	add_unistate("st_tex_matrix", ST_MATRIX4);
	//add_unistate("st_tex_matrix_inverse", ST_MATRIX4);

	//add_unistate("st_mvp_matrix", ST_MATRIX4);
	//add_unistate("st_mvp_matrix_inverse", ST_MATRIX4);

	add_unistate("st_fbsize", ST_FLOAT2);

	add_unistate("st_time", ST_FLOAT);

	add_unistate("st_light_pos", ST_FLOAT3);
	add_unistate("st_light_color", ST_FLOAT3);
	add_unistate("st_light_radius", ST_FLOAT);

	add_unistate("st_mtl_diffuse", ST_FLOAT3);
	add_unistate("st_mtl_specular", ST_FLOAT3);
	add_unistate("st_mtl_shininess", ST_FLOAT);
	add_unistate("st_mtl_alpha", ST_FLOAT);

	// setup some reasonable default values (everything else will be zero, matrices identity)
	set_unistate("st_light_color", Vector3(1, 1, 1));
	set_unistate("st_mtl_diffuse", Vector3(1, 1, 1));
	set_unistate("st_mtl_alpha", 1.0f);
}

static bool create_screens()
{
	BaseScreen *scr;

	// create the debug overlay screen
	debug_screen = new DebugScreen;
	if(!debug_screen->init()) {
		return false;
	}

	// create the starting screen last
	scr = new GameScreen;
	if(!scr->init()) {
		return false;
	}
	screens.push_back(scr);

	info_log("start-screen: %s\n", scr->get_name());
	push_screen(scr);
	return true;
}

static void toggle_debug()
{
	if(dynamic_cast<DebugScreen*>(current_screen())) {
		pop_screen();
	} else {
		push_screen(debug_screen);
	}
}
