#include <algorithm>
#include <float.h>
#include "geom.h"
#include "logger.h"

GeomObject::~GeomObject()
{
}


Sphere::Sphere()
{
	radius = 1.0;
}

Sphere::Sphere(const Vector3 &cent, float radius)
	: center(cent)
{
	this->radius = radius;
}

void Sphere::set_union(const GeomObject *obj1, const GeomObject *obj2)
{
	const Sphere *sph1 = dynamic_cast<const Sphere*>(obj1);
	const Sphere *sph2 = dynamic_cast<const Sphere*>(obj2);

	if(!sph1 || !sph2) {
		error_log("Sphere::set_union: arguments must be spheres");
		return;
	}

	float dist = (sph1->center - sph2->center).length();
	float surf_dist = dist - (sph1->radius + sph2->radius);
	float d1 = sph1->radius + surf_dist / 2.0;
	float d2 = sph2->radius + surf_dist / 2.0;
	float t = d1 / (d1 + d2);

	if(t < 0.0) t = 0.0;
	if(t > 1.0) t = 1.0;

	center = sph1->center * t + sph2->center * (1.0 - t);
	radius = std::max(dist * t + sph2->radius, dist * (1.0f - t) + sph1->radius);
}

void Sphere::set_intersection(const GeomObject *obj1, const GeomObject *obj2)
{
	error_log("Sphere::intersection undefined\n");
}

bool Sphere::intersect(const Ray &ray, HitPoint *hit) const
{
	float a = dot_product(ray.dir, ray.dir);
	float b = 2.0 * ray.dir.x * (ray.origin.x - center.x) +
		2.0 * ray.dir.y * (ray.origin.y - center.y) +
		2.0 * ray.dir.z * (ray.origin.z - center.z);
	float c = dot_product(ray.origin, ray.origin) + dot_product(center, center) -
		2.0 * dot_product(ray.origin, center) - radius * radius;

	float discr = b * b - 4.0 * a * c;
	if(discr < 1e-4) {
		return false;
	}

	float sqrt_discr = sqrt(discr);
	float t0 = (-b + sqrt_discr) / (2.0 * a);
	float t1 = (-b - sqrt_discr) / (2.0 * a);

	if(t0 < 1e-4)
		t0 = t1;
	if(t1 < 1e-4)
		t1 = t0;

	float t = t0 < t1 ? t0 : t1;
	if(t < 1e-4) {
		return false;
	}

	// fill the HitPoint structure
	if(hit) {
		hit->obj = this;
		hit->dist = t;
		hit->pos = ray.origin + ray.dir * t;
		hit->normal = (hit->pos - center) / radius;
	}
	return true;
}

bool Sphere::contains(const Vector3 &pt) const
{
	return (pt - center).length_sq() <= radius * radius;
}

float Sphere::distance(const Vector3 &v) const
{
	return (v - center).length() - radius;
}


Cone::Cone()
	: dir(0, -1, 0)
{
	set_angle(M_PI / 4.0);
	begin = 0.0;
	end = 1.0;
}

Cone::Cone(const Vector3 &aapex, const Vector3 &adir)
	: apex(aapex), dir(adir)
{
	set_angle(M_PI / 4.0);
	begin = 0.0;
	end = 1.0;
}

Cone::Cone(const Vector3 &aapex, const Vector3 &adir, float angle)
	: apex(aapex), dir(adir)
{
	set_angle(angle);
	begin = 0.0;
	end = 1.0;
}

void Cone::set_angle(float theta)
{
	angle = theta;
	slope = tan(theta);
}

void Cone::set_angle_deg(float theta)
{
	angle = DEG_TO_RAD(theta);
	slope = tan(angle);
}

float Cone::get_angle() const
{
	return angle;
}

float Cone::get_angle_deg() const
{
	return RAD_TO_DEG(angle);
}

float Cone::get_slope() const
{
	return slope;
}

void Cone::calc_local_matrix() const
{
	Vector3 vj = dir.normalized();
	Vector3 vk = fabs(dir.z) < 0.99 ? Vector3(0, 0, 1) : Vector3(1, 0, 0);
	Vector3 vi = cross_product(vj, vk).normalized();
	vk = cross_product(vi, vj).normalized();

	Matrix4x4 rot;
	rot.set_column_vector(vi, 0);
	rot.set_column_vector(vj, 1);
	rot.set_column_vector(vk, 2);
	rot.set_row_vector(Vector4(0, 0, 0, 1), 3);
	xform_dir = rot;

	Matrix4x4 trans;
	trans.set_translation(apex);
	xform = trans * rot;

	trans.set_translation(-apex);
	inv_xform_dir = rot.transposed();
	inv_xform = inv_xform_dir * trans;
}

bool Cone::intersect(const Ray &inray, HitPoint *hit) const
{
	float ymin = std::min(begin, end);
	//float ymax = std::max(begin, end);

	calc_local_matrix();
	Ray ray = inray.transformed(inv_xform);

	// find intersections at the cone body
	HitPoint hit_cone;
	bool res_cone = intersect_nocaps_local(ray, &hit_cone);

	// find intersections at the caps
	HitPoint hit_caps;
	Plane cap;
	cap.pt = Vector3(0, ymin, 0);
	cap.normal = Vector3(0, -1, 0);
	bool res_caps = cap.intersect(ray, &hit_caps);
	if(res_caps) {
		float rad_at_cap = slope * ymin;
		float hit_rad_sq = (hit_caps.pos - cap.pt).length_sq();
		if(hit_rad_sq >= rad_at_cap * rad_at_cap) {
			res_caps = false;
		}
	}

	// return the nearest intersection (if any)
	if(res_caps && res_cone) {
		*hit = hit_cone.dist <= hit_caps.dist ? hit_cone : hit_caps;
	} else if(res_caps) {
		*hit = hit_caps;
	} else if(res_cone) {
		*hit = hit_cone;
	}

	if(res_caps || res_cone) {
		hit->obj = this;
		hit->pos.transform(xform);
		hit->normal.transform(xform_dir);
		return true;
	}
	return false;
}

#define SQ(x)	((x) * (x))
bool Cone::intersect_nocaps_local(const Ray &ray, HitPoint *hit) const
{
	float ymin = std::min(begin, end);
	float ymax = std::max(begin, end);

	float a = SQ(ray.dir.x) - SQ(slope * ray.dir.y) + SQ(ray.dir.z);
	float b = 2.0 * ray.origin.x * ray.dir.x - 2.0 * SQ(slope) * ray.origin.y * ray.dir.y +
		2.0 * ray.origin.z * ray.dir.z;
	float c = SQ(ray.origin.x) - SQ(slope * ray.origin.y) + SQ(ray.origin.z);

	float discr = b * b - 4.0 * a * c;
	if(discr < 1e-4)
		return false;

	float sqrt_discr = sqrt(discr);
	float t0 = (-b + sqrt_discr) / (2.0 * a);
	float t1 = (-b - sqrt_discr) / (2.0 * a);

	float y0 = ray.origin.y + ray.dir.y * t0;
	float y1 = ray.origin.y + ray.dir.y * t1;

	if(t0 < 1e-4 || y0 < ymin || y0 >= ymax) {
		t0 = t1;
	}
	if(t1 < 1e-4 || y1 < ymin || y1 >= ymax) {
		t1 = t0;
	}

	float t = t0 < t1 ? t0 : t1;
	if(t < 1e-4)
		return false;

	// fill the HitPoint structure
	hit->dist = t;
	hit->pos = ray.origin + ray.dir * t;

	if(hit->pos.y < ymin || hit->pos.y >= ymax) {
		return false;
	}

	float radius = hit->pos.y * slope;
	hit->normal = Vector3(hit->pos.x, 0.0, hit->pos.z) / radius;

	float ny = -sin(angle);
	float xzlen = sqrt(1.0 - ny * ny);
	hit->normal.x *= xzlen;
	hit->normal.y = ny;
	hit->normal.z *= xzlen;

	hit->obj = this;
	return true;
}

// TODO implement
bool Cone::sweep_intersect(const Ray &ray, float rad, HitPoint *hit) const
{
	if(intersect(ray, hit)) {
		return true;
	}
	return false;
}

bool Cone::contains(const Vector3 &pt) const
{
	Vector3 v = pt - apex;
	float d = v.length();
	if(d == 0.0) {
		return begin <= 0.0; // avoid divby0 down the road
	}

	float projlen = dot_product(v, dir.normalized());
	float costheta = projlen / d;

	if(costheta < cos(angle)) {
		return false;
	}

	return projlen >= begin && projlen <= end;
}


float Cone::distance(const Vector3 &p) const
{
	calc_local_matrix();

	Vector2 side_dir = Vector2(sin(angle), cos(angle));

	Vector3 ploc = p.transformed(inv_xform);
	Vector2 norm2 = Vector2(side_dir.y, -side_dir.x);
	Vector2 p2 = Vector2(sqrt(ploc.x * ploc.x + ploc.z * ploc.z), ploc.y);

	float tan_theta = tan(angle);
	float rtop = begin * tan_theta;
	float rbot = end * tan_theta;

	float top = begin;
	float bot = end;

	// above the top cap ...
	if(p2.y < top) {
		// and inside the cap radius
		if(p2.x < rtop) {
			return top - p2.y;
		}
		// otherwise it's just the distance from the top-most point of the side
		return (p2 - Vector2(rtop, top)).length();
	}
	// below the bottom cap ...
	if(p2.y > bot) {
		// and inside the cap radius
		if(p2.x < rbot) {
			return p2.y - bot;
		}
		// otherwise it's just the distance from the bottom-most point of the side
		return (p2 - Vector2(rbot, bot)).length();
	}

	// we're between caps vertically
	float sdist = dot_product(norm2, p2);
	if(sdist <= 0.0) {
		// if we're inside, then find the minimum abs dist of caps,side
		return -std::min(-sdist, std::min(p2.y - top, bot - p2.y));
	}

	// we're outside, so it's the minimum between the side dist, and bottom point
	float blen_sq = (p2 - Vector2(rbot, bot)).length_sq();
	return blen_sq < sdist * sdist ? sqrt(blen_sq) : sdist;
}


AABox::AABox()
{
}

AABox::AABox(const Vector3 &vmin, const Vector3 &vmax)
	: min(vmin), max(vmax)
{
}

void AABox::set_union(const GeomObject *obj1, const GeomObject *obj2)
{
	const AABox *box1 = dynamic_cast<const AABox*>(obj1);
	const AABox *box2 = dynamic_cast<const AABox*>(obj2);

	if(!box1 || !box2) {
		error_log("AABox::set_union: arguments must be AABoxes too\n");
		return;
	}

	min.x = std::min(box1->min.x, box2->min.x);
	min.y = std::min(box1->min.y, box2->min.y);
	min.z = std::min(box1->min.z, box2->min.z);

	max.x = std::max(box1->max.x, box2->max.x);
	max.y = std::max(box1->max.y, box2->max.y);
	max.z = std::max(box1->max.z, box2->max.z);
}

void AABox::set_intersection(const GeomObject *obj1, const GeomObject *obj2)
{
	const AABox *box1 = dynamic_cast<const AABox*>(obj1);
	const AABox *box2 = dynamic_cast<const AABox*>(obj2);

	if(!box1 || !box2) {
		error_log("AABox::set_intersection: arguments must be AABoxes too\n");
		return;
	}

	for(int i=0; i<3; i++) {
		min[i] = std::max(box1->min[i], box2->min[i]);
		max[i] = std::min(box1->max[i], box2->max[i]);

		if(max[i] < min[i]) {
			max[i] = min[i];
		}
	}
}

bool AABox::intersect(const Ray &ray, HitPoint *hit) const
{
	Vector3 param[2] = {min, max};
	Vector3 inv_dir(1.0 / ray.dir.x, 1.0 / ray.dir.y, 1.0 / ray.dir.z);
	int sign[3] = {inv_dir.x < 0, inv_dir.y < 0, inv_dir.z < 0};

	float tmin = (param[sign[0]].x - ray.origin.x) * inv_dir.x;
	float tmax = (param[1 - sign[0]].x - ray.origin.x) * inv_dir.x;
	float tymin = (param[sign[1]].y - ray.origin.y) * inv_dir.y;
	float tymax = (param[1 - sign[1]].y - ray.origin.y) * inv_dir.y;

	if(tmin > tymax || tymin > tmax) {
		return false;
	}
	if(tymin > tmin) {
		tmin = tymin;
	}
	if(tymax < tmax) {
		tmax = tymax;
	}

	float tzmin = (param[sign[2]].z - ray.origin.z) * inv_dir.z;
	float tzmax = (param[1 - sign[2]].z - ray.origin.z) * inv_dir.z;

	if(tmin > tzmax || tzmin > tmax) {
		return false;
	}
	if(tzmin > tmin) {
		tmin = tzmin;
	}
	if(tzmax < tmax) {
		tmax = tzmax;
	}

	float t = tmin < 1e-4 ? tmax : tmin;
	if(t >= 1e-4) {

		if(hit) {
			hit->obj = this;
			hit->dist = t;
			hit->pos = ray.origin + ray.dir * t;

			float min_dist = FLT_MAX;
			Vector3 offs = min + (max - min) / 2.0;
			Vector3 local_hit = hit->pos - offs;

			static const Vector3 axis[] = {
				Vector3(1, 0, 0), Vector3(0, 1, 0), Vector3(0, 0, 1)
			};
			//int tcidx[][2] = {{2, 1}, {0, 2}, {0, 1}};

			for(int i=0; i<3; i++) {
				float dist = fabs((max[i] - offs[i]) - fabs(local_hit[i]));
				if(dist < min_dist) {
					min_dist = dist;
					hit->normal = axis[i] * (local_hit[i] < 0.0 ? 1.0 : -1.0);
					//hit->texcoord = Vector2(hit->pos[tcidx[i][0]], hit->pos[tcidx[i][1]]);
				}
			}
		}
		return true;
	}
	return false;
}

bool AABox::contains(const Vector3 &v) const
{
	return v.x >= min.x && v.y >= min.y && v.z >= min.z &&
		v.x <= max.x && v.y <= max.y && v.z <= max.z;
}

float AABox::distance(const Vector3 &v) const
{
	return 0.0;	// TODO
}


Plane::Plane()
	: normal(0.0, 1.0, 0.0)
{
}

Plane::Plane(const Vector3 &p, const Vector3 &norm)
	: pt(p)
{
	normal = norm.normalized();
}

Plane::Plane(const Vector3 &p1, const Vector3 &p2, const Vector3 &p3)
	: pt(p1)
{
	normal = cross_product(p2 - p1, p3 - p1).normalized();
}

Plane::Plane(const Vector3 &normal, float dist)
{
	this->normal = normal.normalized();
	pt = this->normal * dist;
}

void Plane::set_union(const GeomObject *obj1, const GeomObject *obj2)
{
	error_log("Plane::set_union undefined\n");
}

void Plane::set_intersection(const GeomObject *obj1, const GeomObject *obj2)
{
	error_log("Plane::set_intersection undefined\n");
}

bool Plane::intersect(const Ray &ray, HitPoint *hit) const
{
	float ndotdir = dot_product(normal, ray.dir);
	if(fabs(ndotdir) < 1e-4) {
		return false;
	}

	if(hit) {
		Vector3 ptdir = pt - ray.origin;
		float t = dot_product(normal, ptdir) / ndotdir;

		hit->dist = t;
		hit->pos = ray.origin + ray.dir * t;
		hit->normal = normal;
		hit->obj = this;
	}
	return true;
}

bool Plane::contains(const Vector3 &v) const
{
	return dot_product(v, normal) <= 0.0;
}

float Plane::distance(const Vector3 &v) const
{
	return dot_product(v - pt, normal);
}
