#ifndef GEOMOBJ_H_
#define GEOMOBJ_H_

#include "vmath/vmath.h"

class GeomObject;

struct HitPoint {
	float dist;				//< parametric distance along the ray
	Vector3 pos;			//< position of intersection (orig + dir * dist)
	Vector3 normal;			//< normal at the point of intersection
	const void *obj;		//< pointer to the intersected object
};

class GeomObject {
public:
	virtual ~GeomObject();

	virtual void set_union(const GeomObject *obj1, const GeomObject *obj2) = 0;
	virtual void set_intersection(const GeomObject *obj1, const GeomObject *obj2) = 0;

	virtual bool intersect(const Ray &ray, HitPoint *hit = 0) const = 0;
	virtual bool contains(const Vector3 &pt) const = 0;

	virtual float distance(const Vector3 &v) const = 0;
};

class Sphere : public GeomObject {
public:
	Vector3 center;
	float radius;

	Sphere();
	Sphere(const Vector3 &center, float radius);

	void set_union(const GeomObject *obj1, const GeomObject *obj2);
	void set_intersection(const GeomObject *obj1, const GeomObject *obj2);

	bool intersect(const Ray &ray, HitPoint *hit = 0) const;
	bool contains(const Vector3 &pt) const;

	float distance(const Vector3 &v) const;
};

class Cone : public GeomObject {
protected:
	float angle, slope;	// slope derived from angle

	mutable Matrix4x4 xform, inv_xform;
	mutable Matrix4x4 xform_dir, inv_xform_dir;

	void calc_local_matrix() const;

	bool intersect_nocaps_local(const Ray &ray, HitPoint *hit = 0) const;
public:
	Vector3 apex, dir;
	float begin, end;

	Cone();
	Cone(const Vector3 &apex, const Vector3 &dir);
	Cone(const Vector3 &apex, const Vector3 &dir, float angle);

	void set_angle(float theta);
	void set_angle_deg(float theta);
	float get_angle() const;
	float get_angle_deg() const;

	float get_slope() const;

	// unimplemented for the cone
	void set_union(const GeomObject *obj1, const GeomObject *obj2) {}
	void set_intersection(const GeomObject *obj1, const GeomObject *obj2) {}

	bool intersect(const Ray &ray, HitPoint *hit = 0) const;
	bool sweep_intersect(const Ray &ray, float rad, HitPoint *hit = 0) const;
	bool contains(const Vector3 &pt) const;

	float distance(const Vector3 &v) const;
};

class AABox : public GeomObject {
public:
	Vector3 min, max;

	AABox();
	AABox(const Vector3 &min, const Vector3 &max);

	void set_union(const GeomObject *obj1, const GeomObject *obj2);
	void set_intersection(const GeomObject *obj1, const GeomObject *obj2);

	bool intersect(const Ray &ray, HitPoint *hit = 0) const;
	bool contains(const Vector3 &pt) const;

	float distance(const Vector3 &v) const;
};

class Plane : public GeomObject {
public:
	Vector3 pt, normal;

	Plane();
	Plane(const Vector3 &pt, const Vector3 &normal);
	Plane(const Vector3 &p1, const Vector3 &p2, const Vector3 &p3);
	Plane(const Vector3 &normal, float dist);

	void set_union(const GeomObject *obj1, const GeomObject *obj2);
	void set_intersection(const GeomObject *obj1, const GeomObject *obj2);

	bool intersect(const Ray &ray, HitPoint *hit = 0) const;
	bool contains(const Vector3 &pt) const;

	float distance(const Vector3 &v) const;
};

#endif	// GEOMOBJ_H_
