#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "options.h"
#include "assman.h"
#include "logger.h"

enum {
	OPT_RESOLUTION,
	OPT_FULLSCREEN,
	OPT_WINDOWED,
	OPT_MUSIC_TOGGLE,
	OPT_MUSIC,
	OPT_NOMUSIC,
	OPT_BRANCHES,
	OPT_SMOKE,
	OPT_NOSMOKE,
	OPT_HELP
};
static struct {
	char c;
	const char *s;
	int opt;
	const char *desc;
} optnames[] = {
	{ 's', "size", OPT_RESOLUTION, "set resolution (WxH)" },
	{ 'f', "fullscreen", OPT_FULLSCREEN, "start fullscreen" },
	{ 'w', "windowed", OPT_WINDOWED, "start windowed" },
	{ 'm', "music", OPT_MUSIC, "enable music" },
	{ 'n', "no-music", OPT_NOMUSIC, "disable music" },
	{ 0, "branches", OPT_BRANCHES, "number of branches per level (60,8)" },
	{ 0, "smoke", OPT_SMOKE, "use a smoke particle system" },
	{ 0, "no-smoke", OPT_NOSMOKE, "don't use a smoke particle system" },
	{ 'h', "help", OPT_HELP, "print usage and exit" },
	{ 0, 0, -1 }
};

static int get_option(char c, const char *s);
static char *strip_space(char *s);
static int bool_value(const char *s);
static void print_usage(const char *argv0);

Options opt;

Options::Options()
{
	// defaults
	width = 1024;
	height = 768;
	music = false;
	fullscreen = false;

	nbranches[0] = 60;
	nbranches[1] = 8;
	smoke = true;
}

bool parse_args(int argc, char **argv)
{
	for(int i=1; i<argc; i++) {
		if(argv[i][0] == '-') {
			int o = -1;
			if(argv[i][2] == 0) {
				o = get_option(argv[i][1], 0);
			} else {
				o = get_option(0, argv[i] + 1);
			}

			switch(o) {
			case OPT_RESOLUTION:
				if(sscanf(argv[++i], "%dx%d", &opt.width, &opt.height) != 2) {
					error_log("%s must be followed by a resolution (WxH)\n", argv[i - 1]);
					return false;
				}
				break;

			case OPT_BRANCHES:
				if(sscanf(argv[++i], "%d,%d,%d,%d,%d", opt.nbranches, opt.nbranches + 1,
							opt.nbranches + 2, opt.nbranches + 3, opt.nbranches + 4) < 1) {
					error_log("%s must be followed by a comma-separated list of branching counts per level (60,8)\n", argv[i - 1]);
					return false;
				}
				break;

			case OPT_SMOKE:
				opt.smoke = true;
				break;

			case OPT_NOSMOKE:
				opt.smoke = false;
				break;

			case OPT_FULLSCREEN:
				opt.fullscreen = true;
				break;

			case OPT_WINDOWED:
				opt.fullscreen = false;
				break;

			case OPT_MUSIC_TOGGLE:
				opt.music = !opt.music;
				break;

			case OPT_MUSIC:
				opt.music = true;
				break;

			case OPT_NOMUSIC:
				opt.music = false;
				break;

			case OPT_HELP:
				print_usage(argv[0]);
				exit(0);

			default:
				error_log("invalid option: %s\n", argv[i]);
				return false;
			}
		} else {
			error_log("unexpected argument: %s\n", argv[i]);
			return false;
		}
	}
	return true;
}

#define SKIP_INVAL	\
	warning_log("skipping invalid config line: %s\n", orig); continue

bool load_config(const char *fname)
{
	ass_file *fp;
	char buf[256], orig[256];

	if(!(fp = ass_fopen(fname, "rb"))) {
		warning_log("failed to load config file %s, using defaults\n", fname);
		return false;
	}

	while(ass_fgets(buf, sizeof buf, fp)) {
		strcpy(orig, buf);	// keeping the original line intact for error messages
		char *sep = strchr(buf, '=');
		if(!sep) {
			SKIP_INVAL;
		}
		*sep = 0;

		char *key = strip_space(buf);
		char *val = strip_space(sep + 1);
		int o = get_option(0, key);
		int tmp[8];

		switch(o) {
		case OPT_RESOLUTION:
			if(sscanf(val, "%dx%d", tmp, tmp + 1) != 2) {
				SKIP_INVAL;
			}
			opt.width = tmp[0];
			opt.height = tmp[1];
			break;

		case OPT_BRANCHES:
			memset(tmp, 0, sizeof tmp);
			if(sscanf(val, "%d,%d,%d,%d,%d", tmp, tmp + 1, tmp + 2, tmp + 3, tmp + 4) < 1) {
				SKIP_INVAL;
			}
			for(int i=0; i<5; i++) {
				opt.nbranches[i] = tmp[i];
			}
			break;

		case OPT_SMOKE:
			opt.smoke = bool_value(val);
			break;

		case OPT_FULLSCREEN:
			opt.fullscreen = bool_value(val);
			break;

		case OPT_WINDOWED:
			opt.fullscreen = !bool_value(val);
			break;

		case OPT_MUSIC:
			opt.music = bool_value(val);
			break;

		default:
			SKIP_INVAL;
		}
	}

	ass_fclose(fp);
	return true;
}

static int get_option(char c, const char *s)
{
	for(int i=0; optnames[i].opt != -1; i++) {
		if((c && c == optnames[i].c) || (s && strcmp(s, optnames[i].s) == 0)) {
			return optnames[i].opt;
		}
	}
	return -1;
}

static char *strip_space(char *s)
{
	char *res, *last_nonsp = 0;
	while(*s && isspace(*s)) s++;
	res = s;
	while(*s) {
		if(!isspace(*s)) {
			last_nonsp = s;
		}
		++s;
	}
	if(last_nonsp) last_nonsp[1] = 0;
	return res;
}

static int bool_value(const char *s)
{
	int sz = strlen(s);
	char *buf = (char*)alloca(sz + 1);
	char *dest = buf;

	while(*s) {
		*dest++ = tolower(*s++);
	}
	*dest = 0;

	if(strcmp(buf, "true") == 0 || strcmp(buf, "yes") == 0 || strcmp(buf, "1") == 0) {
		return 1;
	}
	if(strcmp(buf, "false") == 0 || strcmp(buf, "no") == 0 || strcmp(buf, "0") == 0) {
		return 0;
	}
	return -1;
}

static void print_usage(const char *argv0)
{
	printf("Usage: %s [options]\n", argv0);
	printf("Options:\n");

	for(int i=0; optnames[i].opt != -1; i++) {
		putchar(' ');
		if(optnames[i].c) {
			printf("-%c", optnames[i].c);
		}
		if(optnames[i].s) {
			printf("%s-%s", optnames[i].c ? ", " : " ", optnames[i].s);
		}
		printf(": %s\n", optnames[i].desc);
	}
}
