#include <stdlib.h>
#include <map>
#include "opengl.h"
#include "points.h"
#include "sdrman.h"
#include "logger.h"

#define INIT_ONCE \
	do { \
		if(!vbo && !init())	\
			abort(); \
	} while(0)

static bool init();
static void flush();

static Shader *def_vsdr;
static ShaderProg *def_sdrprog;
static ShaderProg *sdrprog;

static std::map<Shader*, ShaderProg*> usdrmap;

#define MAX_POINTS	1024
#define MAX_VERTS	(MAX_POINTS * 4)

static float pos_buf[MAX_VERTS * 4];
static float col_buf[MAX_VERTS * 4];
static int pcount;
static float *pos_ptr, *col_ptr;
static float cur_color[4];

static unsigned int vbo, ibo;	// fixed STATIC_DRAW buffers for vertices and indices
static struct {
	unsigned int vbo_pos;		// streaming point positions * 4
	unsigned int vbo_color;		// streaming point colors * 4
} dynvbo[2];
static int cur_buf;		// current dynamic buffer set

// attribute locations
enum {
	ATTR_VERTEX = 0,	// quad vertices in local space
	ATTR_POS	= 1,	// position in world space
	ATTR_COLOR	= 2
};

bool point_pixel_shader(Shader *sdr)
{
	INIT_ONCE;

	if(!sdr) {
		sdrprog = def_sdrprog;
		return true;
	}

	if((sdrprog = usdrmap[sdr])) {
		return true;
	}
	sdrprog = new ShaderProg;
	if(!sdrprog->create(def_vsdr, sdr)) {
		sdrprog = def_sdrprog;
		return false;
	}

	usdrmap[sdr] = sdrprog;
	return true;
}

void begin_points()
{
	pcount = 0;
	flush();
}

void end_points()
{
	flush();
}

void point_color(float r, float g, float b, float a)
{
	cur_color[0] = r;
	cur_color[1] = g;
	cur_color[2] = b;
	cur_color[3] = a;
}

void draw_point(const float *pos, float size)
{
	draw_point(pos, cur_color, size);
}

void draw_point(const float *pos, const float *color, float size)
{
	for(int i=0; i<4; i++) {	// 4 verts
		for(int j=0; j<4; j++) {	// xyzw
			*pos_ptr++ = j < 3 ? pos[j] : size;	// put the size in w
			*col_ptr++ = color[j];
		}
	}

	if(++pcount >= MAX_POINTS) {
		flush();
	}
}

void draw_points(int num, const float *pos, const float *colors, float *sizes)
{
	for(int i=0; i<num; i++) {
		for(int j=0; j<4; j++) {	// 4 verts per quad
			for(int k=0; k<4; k++) {	// xyzw
				*pos_ptr++ = k < 3 ? pos[k] : *sizes;
				*col_ptr++ = colors[k];
			}
		}
		pos += 3;
		colors += 4;
		++sizes;

		if(++pcount >= MAX_POINTS) {
			flush();
		}
	}
}

static bool init()
{
	// load shaders
	if(!(def_vsdr = vsdrset.get("bb.v.glsl"))) {
		return false;
	}
	Shader *psdr = psdrset.get("bb.p.glsl");
	if(!psdr) {
		return false;
	}
	def_sdrprog = new ShaderProg;
	if(!def_sdrprog->create(def_vsdr, psdr)) {
		return false;
	}
	def_sdrprog->set_attrib_location("a_vertex", ATTR_VERTEX);
	def_sdrprog->set_attrib_location("a_pos", ATTR_POS);
	def_sdrprog->set_attrib_location("a_color", ATTR_COLOR);
	sdrprog = def_sdrprog;

	// create the streaming vertex buffers
	for(int i=0; i<2; i++) {
		glGenBuffers(1, &dynvbo[i].vbo_pos);
		glBindBuffer(GL_ARRAY_BUFFER, dynvbo[i].vbo_pos);
		glBufferData(GL_ARRAY_BUFFER, MAX_VERTS * 4 * sizeof(float), 0, GL_STREAM_DRAW);

		glGenBuffers(1, &dynvbo[i].vbo_color);
		glBindBuffer(GL_ARRAY_BUFFER, dynvbo[i].vbo_color);
		glBufferData(GL_ARRAY_BUFFER, MAX_VERTS * 4 * sizeof(float), 0, GL_STREAM_DRAW);
	}

	// create the fixed vertex array and fill it with data
	static const float vpat[] = {-1, -1, 0, 1, -1, 0, 1, 1, 0, -1, 1, 0};
	float *verts, *vptr;
	try {
		vptr = verts = new float[MAX_VERTS * 4];
	}
	catch(...) {
		return false;
	}

	for(int i=0; i<MAX_POINTS; i++) {
		for(int j=0; j<12; j++) {	// 4 verts per quad, 3 floats per vert
			*vptr++ = vpat[j];
		}
	}

	glGenBuffers(1, &vbo);
	glBindBuffer(GL_ARRAY_BUFFER, vbo);
	glBufferData(GL_ARRAY_BUFFER, MAX_VERTS * 3 * sizeof(float), verts, GL_STATIC_DRAW);
	delete [] verts;

	// create the static index buffer and fill it with data
	static const unsigned int idxpat[] = {0, 1, 2, 0, 2, 3};
	unsigned int *index, *iptr;
	try {
		iptr = index = new unsigned int[MAX_POINTS * 6];
	}
	catch(...) {
		return false;
	}

	for(int i=0; i<MAX_POINTS; i++) {
		for(int j=0; j<6; j++) {
			*iptr++ = i * 4 + idxpat[j];
		}
	}

	glGenBuffers(1, &ibo);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, MAX_POINTS * 6 * sizeof(unsigned int), index, GL_STATIC_DRAW);
	delete [] index;

	return true;
}

static void flush()
{
	if(pcount) {
		int num_verts = pcount * 4;

		INIT_ONCE;

		sdrprog->bind();

		glEnableVertexAttribArray(ATTR_VERTEX);
		glEnableVertexAttribArray(ATTR_POS);
		glEnableVertexAttribArray(ATTR_COLOR);

		// fixed local space vertex buffer
		glBindBuffer(GL_ARRAY_BUFFER, vbo);
		glVertexAttribPointer(ATTR_VERTEX, 3, GL_FLOAT, 0, 0, 0);

		// update and use the position buffer
		glBindBuffer(GL_ARRAY_BUFFER, dynvbo[cur_buf].vbo_pos);
		glBufferSubData(GL_ARRAY_BUFFER, 0, num_verts * 4 * sizeof(float), pos_buf);
		glVertexAttribPointer(ATTR_POS, 4, GL_FLOAT, 0, 0, 0);

		// update and use the color buffer
		glBindBuffer(GL_ARRAY_BUFFER, dynvbo[cur_buf].vbo_color);
		glBufferSubData(GL_ARRAY_BUFFER, 0, num_verts * 4 * sizeof(float), col_buf);
		glVertexAttribPointer(ATTR_COLOR, 4, GL_FLOAT, 1, 0, 0);

		glBindBuffer(GL_ARRAY_BUFFER, 0);

		// draw using the fixed index buffer
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, ibo);
		glDrawElements(GL_TRIANGLES, pcount * 6, GL_UNSIGNED_INT, 0);
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);

		glDisableVertexAttribArray(ATTR_VERTEX);
		glDisableVertexAttribArray(ATTR_POS);
		glDisableVertexAttribArray(ATTR_COLOR);
	}

	pcount = 0;
	pos_ptr = pos_buf;
	col_ptr = col_buf;
	cur_buf = (cur_buf + 1) % 2;
}
