#include "psys_fire.h"
#include "xtree.h"
#include "rng.h"
#include "logger.h"


FireParticleSystem::FireParticleSystem()
{
	tree = 0;
	orig_spawn_rate = -1;
}

void FireParticleSystem::set_parent_tree(PineTree *tree)
{
	this->tree = tree;
}

bool FireParticleSystem::spawn(struct psys_particle *p) const
{
	if(spawn_cdf.empty()) {
		return false;
	}

// pick a branch based on the cumulative probabilities in spawn_cdf
	PineTreeBranch *br = 0;
	float rndvar = rng_frand();
	for(size_t i=0; i<spawn_cdf.size(); i++) {
		if(rndvar <= spawn_cdf[i].prob) {
			br = spawn_cdf[i].branch;
			break;
		}
	}

	if(!br) {
		warning_log("failed to pick a branch!\n");
		return false;
	}

	// pick a spot on the branch within the burn intervals
	rndvar = rng_frand();
	float t0, t1;
	if(!br->get_burn_fronts(&t0, &t1)) {
		abort();
	}

	float t = lerp(t0, t1, rndvar);
	Vector3 pos = lerp(br->end[0], br->end[1], t);

	p->pos.x = pos.x;
	p->pos.y = pos.y;
	p->pos.z = pos.z;
	return true;
}

void FireParticleSystem::update(long tm)
{
	if(!tree) return;

	// prepare for particle spawning by constructing the cumulative probability
	// table based on burning branch intervals
	float total = 0.0f;
	std::list<PineTreeBranch*>::const_iterator it = tree->burning_branches.begin();
	while(it != tree->burning_branches.end()) {
		PineTreeBranch *br = *it++;

		if(br->burned <= 0.0f || br->burned >= 1.0f) {
			continue;
		}

		total += br->burned * br->size;

		SpawnCDFEntry entry;
		entry.prob = total;
		entry.branch = br;

		spawn_cdf.push_back(entry);
	}

	// normalize the cdf
	if(total > 0.0f) {
		float one_over_total = 1.0f / total;
		for(size_t i=0; i<spawn_cdf.size(); i++) {
			spawn_cdf[i].prob *= one_over_total;
		}
	}

	// set the spawn rate as a function of the burn area
	if(orig_spawn_rate < 0.0) {
		orig_spawn_rate = get_attr()->get_spawn_rate();
	}

	float spawn_rate = total * orig_spawn_rate;
	get_attr()->set_spawn_rate(spawn_rate);
	//debug_log("setting spawn rate: %g /s\n", spawn_rate);

	ParticleSystem::update(tm);

	spawn_cdf.clear();
}

