#include <limits.h>
#include "psyspp.h"
#include "sdrman.h"
#include "logger.h"
#include "mesh.h"	// just for the attrib enums
#include "unistate.h"
#include "datapath.h"
#include "texman.h"
#include "texgen.h"
#include "image.h"
#include "points.h"

static int pspawn(struct psys_emitter *ps, struct psys_particle *p, void *cls);

static void pdraw_start(const psys_emitter *em, void *cls);
static void pdraw(const psys_emitter *em, const psys_particle *part, void *cls);
static void pdraw_end(const psys_emitter *em, void *cls);

static unsigned int psys_load_texture(const char *fname, void *cls);

ParticleSystemAttributes::ParticleSystemAttributes()
{
	tex = 0;
	psdr = 0;
	own_psattr = true;
	psattr = new psys_attributes;
	psys_init_attr(psattr);
}

ParticleSystemAttributes::ParticleSystemAttributes(psys_attributes *psattr)
{
	tex = 0;
	psdr = 0;
	own_psattr = false;
	this->psattr = psattr;
}

ParticleSystemAttributes::~ParticleSystemAttributes()
{
	if(own_psattr) {
		psys_destroy_attr(psattr);
		delete psattr;
	}
}

ParticleSystemAttributes::ParticleSystemAttributes(const ParticleSystemAttributes &rhs)
{
	own_psattr = true;
	tex = rhs.tex;
	psdr = rhs.psdr;
	psattr = new psys_attributes;
	psys_init_attr(psattr);

	psys_copy_attr(psattr, rhs.psattr);
}

ParticleSystemAttributes &ParticleSystemAttributes::operator =(const ParticleSystemAttributes &rhs)
{
	if(&rhs != this) {
		tex = rhs.tex;
		psdr = rhs.psdr;
		psys_copy_attr(psattr, rhs.psattr);
	}
	return *this;
}

bool ParticleSystemAttributes::load(const char *fname)
{
	psys_texture_loader(psys_load_texture, 0, this);
	return psys_load_attr(psattr, datafile_path(fname).c_str()) != -1;
}

bool ParticleSystemAttributes::load(ass_file *fp)
{
	psys_texture_loader(psys_load_texture, 0, this);
	return psys_load_attr_stream(psattr, fp) != -1;
}

bool ParticleSystemAttributes::save(const char *fname) const
{
	return psys_save_attr(psattr, fname) != -1;
}

bool ParticleSystemAttributes::save(FILE *fp) const
{
	return psys_save_attr_stream(psattr, fp) != -1;
}

void ParticleSystemAttributes::set_part_color(const Vector3 &color, float t)
{
	psys_set_value3(&psattr->part_attr.color, (anm_time_t)(t * 1000.0), v3_cons(color.x, color.y, color.z));
}

void ParticleSystemAttributes::set_part_alpha(float alpha, float t)
{
	psys_set_value(&psattr->part_attr.alpha, (anm_time_t)(t * 1000.0), alpha);
}

void ParticleSystemAttributes::set_part_scale(float size, float t)
{
	psys_set_value(&psattr->part_attr.size, (anm_time_t)(t * 1000.0), size);
}


// emmiter attributes
void ParticleSystemAttributes::set_shader(Shader *sdr)
{
	psdr = sdr;
}

Shader *ParticleSystemAttributes::get_shader() const
{
	return psdr;
}

void ParticleSystemAttributes::set_texture(Texture *tex)
{
	this->tex = tex;
	psattr->tex = tex->get_id();
}

Texture *ParticleSystemAttributes::get_texture() const
{
	return tex;
}

void ParticleSystemAttributes::set_spawn_range(const Vector3 &range, long tm)
{
	psys_set_value3(&psattr->spawn_range, ANM_MSEC2TM(tm), v3_cons(range.x, range.y, range.z));
}

void ParticleSystemAttributes::set_spawn_rate(float rate, long tm)
{
	psys_set_value(&psattr->rate, ANM_MSEC2TM(tm), rate);
}

float ParticleSystemAttributes::get_spawn_rate(long tm) const
{
	return psys_get_value(&psattr->rate, ANM_MSEC2TM(tm));
}

void ParticleSystemAttributes::set_life(float life, float range, long tm)
{
	psys_set_anm_rnd(&psattr->life, ANM_MSEC2TM(tm), life, range);
}

void ParticleSystemAttributes::set_size(float sz, float range, long tm)
{
	psys_set_anm_rnd(&psattr->size, ANM_MSEC2TM(tm), sz, range);
}

void ParticleSystemAttributes::set_spawn_dir(const Vector3 &dir, const Vector3 &range, long tm)
{
	psys_set_anm_rnd3(&psattr->dir, ANM_MSEC2TM(tm), v3_cons(dir.x, dir.y, dir.z), v3_cons(range.x, range.y, range.z));
}

void ParticleSystemAttributes::set_gravity(const Vector3 &grav, long tm)
{
	psys_set_value3(&psattr->grav, ANM_MSEC2TM(tm), v3_cons(grav.x, grav.y, grav.z));
}

void ParticleSystemAttributes::set_drag(float drag)
{
	psattr->drag = drag;
}

void ParticleSystemAttributes::set_particle_limit(int lim)
{
	psattr->max_particles = lim;
}


// ---- ParticleSystem ----

ParticleSystem::ParticleSystem()
	: attr(&psys.attr)
{
	psys_init(&psys);
	psys_spawn_func(&psys, pspawn, (void*)this);
	psys_draw_func(&psys, pdraw, pdraw_start, pdraw_end, (void*)this);
	start_time = LONG_MIN;
	last_upd_time = LONG_MIN;
}

ParticleSystem::~ParticleSystem()
{
	psys_destroy(&psys);
}

void ParticleSystem::set_start_time(long tm)
{
	start_time = tm;
}

bool ParticleSystem::is_active() const
{
	if(last_upd_time == LONG_MIN) {
		// this emitter is brand-new, never even updated yet, so it's active
		return true;
	}
	float rate = attr.get_spawn_rate(last_upd_time);
	return psys.pcount > 0 || rate > 0.0;
}

void ParticleSystem::stop()
{
	attr.set_spawn_rate(0);
}

ParticleSystemAttributes *ParticleSystem::get_attr()
{
	return &attr;
}

const ParticleSystemAttributes *ParticleSystem::get_attr() const
{
	return &attr;
}

void ParticleSystem::set_attr(const ParticleSystemAttributes &pattr)
{
	attr = pattr;
}

bool ParticleSystem::load(const char *fname)
{
	psys_texture_loader(psys_load_texture, 0, &attr);
	return attr.load(fname);
}

bool ParticleSystem::save(const char *fname) const
{
	return attr.save(fname);
}

bool ParticleSystem::spawn(struct psys_particle *p) const
{
	return true;
}

void ParticleSystem::update(long tm)
{
	if(start_time > LONG_MIN) {
		tm -= start_time;
	}

	Matrix4x4 xform;
	get_xform(tm, &xform);

	Vector3 pos = Vector3(0, 0, 0).transformed(xform);

	psys_set_pos(&psys, v3_cons(pos.x, pos.y, pos.z), 0);
	psys_update(&psys, (double)tm / 1000.0);

	last_upd_time = tm;
}

void ParticleSystem::draw() const
{
	psys_draw(&psys);
}

// ---- custom spawn function ----
static int pspawn(struct psys_emitter *em, struct psys_particle *p, void *cls)
{
	ParticleSystem *ps = (ParticleSystem*)cls;
	if(!ps->spawn(p)) {
		return -1;
	}
	return 0;
}

// ---- particle drawing ----

static void pdraw_start(const psys_emitter *em, void *cls)
{
	ParticleSystem *ps = (ParticleSystem*)cls;

	Texture *tex = ps->get_attr()->get_texture();
	if(tex) {
		tex->bind();
	}

	glEnable(GL_BLEND);
	if(em->attr.blending == PSYS_ALPHA) {
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	} else {
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	}
	glDepthMask(0);

	point_pixel_shader(ps->get_attr()->get_shader());
	begin_points();
}

static void pdraw(const psys_emitter *em, const psys_particle *part, void *cls)
{
	Vector4 color = Vector4(part->color.x, part->color.y, part->color.z, part->alpha);
	draw_point(&part->pos.x, &color.x, part->size);
}

static void pdraw_end(const psys_emitter *em, void *cls)
{
	end_points();
	point_pixel_shader(0);

	glDepthMask(1);
	glDisable(GL_BLEND);
}

static unsigned int psys_load_texture(const char *fname, void *cls)
{
	ParticleSystemAttributes *attr = (ParticleSystemAttributes*)cls;

	Texture *tex = texset.get(fname);
	if(tex) {
		attr->set_texture(tex);
		return tex->get_id();
	}
	return 0;
}
