#include <algorithm>
#include <float.h>
#include "raytracer.h"
#include "image.h"

// define this to test distance functions
#undef DISTFIELD

static PineTree *xtree;
static Image fbimg;
static Matrix4x4 view_matrix;

static Ray calc_primary_ray(int x, int y, int xsz, int ysz);
#ifndef DISTFIELD
static Vector4 raytrace(const Ray &ray);
#else
static Vector4 raymarch(const Ray &ray);
#endif

void rt_resize(int x, int y)
{
	fbimg.create(x, y, Image::FMT_RGBA);
}

void rt_view_matrix(const Matrix4x4 &mat)
{
	view_matrix = mat.inverse();
}


void rt_scene_tree(PineTree *pt)
{
	xtree = pt;
}


void rt_render()
{
	int xsz = fbimg.get_width();
	int ysz = fbimg.get_height();

	unsigned char *pptr = (unsigned char*)fbimg.get_pixels();

	for(int i=0; i<ysz; i++) {
#pragma omp parallel for
		for(int j=0; j<xsz; j++) {
			Ray ray = calc_primary_ray(j, i, xsz, ysz);
			ray.dir *= 256.0f;
			ray.transform(view_matrix);

#ifndef DISTFIELD
			Vector4 color = raytrace(ray);
#else
			Vector4 color = raymarch(ray);
#endif

			int r = (int)(color.x * 255.0);
			int g = (int)(color.y * 255.0);
			int b = (int)(color.z * 255.0);
			int a = (int)(color.w * 255.0);
			pptr[j * 4] = r > 255 ? 255 : r;
			pptr[j * 4 + 1] = g > 255 ? 255 : g;
			pptr[j * 4 + 2] = b > 255 ? 255 : b;
			pptr[j * 4 + 3] = a > 255 ? 255 : a;
		}

		pptr += xsz * 4;
	}
}

Image *rt_get_image()
{
	return &fbimg;
}

static Ray calc_primary_ray(int x, int y, int xsz, int ysz)
{
	const float vfov = DEG_TO_RAD(45.0f);

	float u = (float)x / (float)xsz;
	float v = (float)y / (float)ysz;
	float aspect = (float)xsz / (float)ysz;

	Vector3 dir;
	dir.x = (u * 2.0 - 1.0) * aspect;
	dir.y = 1.0 - v * 2.0;
	dir.z = -1.0 / tan(vfov * 0.5);

	return Ray(Vector3(0, 0, 0), dir);
}

#ifndef DISTFIELD
static Vector4 raytrace(const Ray &ray)
{
	HitPoint pt;
	if(!xtree->intersect(ray, &pt)) {
		return Vector4(0, 0, 0, 0);
	}

	float d = fabs(xtree->root_branch->bcone.distance(pt.pos));
	float t = std::min(d, 1.0f);

	Vector3 color = lerp(Vector3(1.0, 0.0, 0.0), Vector3(0.0, 0.0, 1.0), t);

	//Vector3 color = pt.normal * 0.5 + Vector3(0.5, 0.5, 0.5);
	return Vector4(color.x, color.y, color.z, 1.0);
}

#else	// defined DISTFIELD
static Vector3 calc_grad(const Vector3 &pos, float val)
{
	return Vector3(xtree->distance(pos + Vector3(0.01, 0, 0)) - val,
			xtree->distance(pos + Vector3(0, 0.01, 0)) - val,
			xtree->distance(pos + Vector3(0, 0, 0.01)) - val);
}

static Vector4 raymarch(const Ray &ray)
{
	const float dthres = 0.001;
	const int max_iter = 64;

	Vector3 pt = ray.origin;
	Vector3 dir = ray.dir.normalized();
	Vector3 normal;

	bool hit = false;
	float last_dist = FLT_MAX;

	for(int i=0; i<max_iter; i++) {
		float d = xtree->distance(pt);
		if(d > last_dist) break;
		if(d <= dthres) {
			normal = calc_grad(pt, d);
			hit = true;
			break;
		}

		pt = pt + dir * d;
		last_dist = d;
	}

	if(!hit) {
		return Vector4(0, 0, 0, 0);
	}

	normal.normalize();
	Vector3 color = normal * 0.5 + Vector3(0.5, 0.5, 0.5);
	return Vector4(color.x, color.y, color.z, 1.0);
}
#endif	// DISTFIELD
