#include <assert.h>
#include <algorithm>
#include <list>
#include "scr_game.h"
#include "object.h"
#include "game.h"
#include "input.h"
#include "meshgen.h"
#include "opengl.h"
#include "sdrman.h"
#include "texman.h"
#include "unistate.h"
#include "gfxutil.h"
#include "xtree.h"
#include "logger.h"
#include "drawtext/drawtext.h"
#include "rng.h"
#include "points.h"
#include "audio/ovstream.h"
#include "options.h"
#include "psman.h"
#include "raytracer.h"

struct Snowball {
	float life;
	float size;
	Vector3 pos, vel;
	long hit_time;	/* keep alive and wet for a short time after a hit */

	ParticleSystem *psys_trail;

	Snowball()
	{
		life = 0.0f;
		size = 1.0f;
		psys_trail = 0;
		hit_time = 0;
	}
};

static void draw_snowballs();
static void throw_snow(int x, int y, float snowball_size);
static Ray calc_pick_ray(int x, int y);
static float ground_heightmap(float u, float v, void *cls);

bool dbg_switch;
int dbg_iter;
Ray dbg_ray;
HitPoint dbg_hitpt;
int dbg_rt_pending;
int dbg_show_rt;
Texture *dbg_rt_tex;

static Sphere dbg_last_sball_hit = Sphere(Vector3(0, -100, 0), 1.0);
static bool dbg_show_last_sball_hit;

#define TREE_SIZE	5.0
#define SNOWBALL_SIZE_MIN		0.1
#define SNOWBALL_SIZE_MAX		0.5
#define SNOWBALL_HIT_SCALE	4.0

static int win_width, win_height;
static bool bnstate[32];
static int press_x[32], press_y[32];
static int prev_x, prev_y;
static int abs_dx, abs_dy;
static float cam_theta, cam_phi = 0, cam_dist = 8;
static Vector3 cam_pos = Vector3(0, TREE_SIZE * 0.5, 0);
static PineTree *xtree;
static Object *floor_obj;

static dtx_font *font;
#define FONT_SIZE	20
static ShaderProg *font_sdr, *ground_sdr;
static Shader *snowball_psdr;
static int font_sdr_vattr, font_sdr_tcattr;
static Texture *snowball_tex;

static std::list<Snowball> snowballs;
static unsigned long last_throw_time;
static unsigned long last_upd_time;
static float snowball_size = 0.1;
static Mesh *sball_mesh;

static PSysManager psman;
static ParticleSystemAttributes psattr_snow_trail, psattr_snow_hit;
static ParticleSystem *ps_snowing;

static bool pause_updates;

static AudioStream *music;

static Vector4 bg_horizon_color, bg_zenith_color;


const char *GameScreen::get_name() const
{
	return "game";
}

bool GameScreen::init()
{
	// ---- text rendering ----
	if(!(font = dtx_open_font_glyphmap(datafile_path("sans20.glyphmap").c_str()))) {
		error_log("failed to open glyphmap: sans20.glyphmap\n");
		return false;
	}
	if(!(font_sdr = get_sdrprog("font.v.glsl", "font.p.glsl"))) {
		return false;
	}
	font_sdr_vattr = font_sdr->get_attrib_location("attr_vertex");
	font_sdr_tcattr = font_sdr->get_attrib_location("attr_texcoord");
	assert(font_sdr_vattr != -1 && font_sdr_tcattr != -1);

	// ---- lighting -----
	Vector3 ambient = Vector3(0.1, 0.1, 0.13);
	set_unistate("st_ambient", ambient);

	Vector3 fog_color = Vector3(0.18, 0.13, 0.23);
	glClearColor(fog_color.x, fog_color.y, fog_color.z, 1);
	set_unistate("st_fog_color", fog_color);

	bg_horizon_color = fog_color;
	bg_zenith_color = ambient;

	set_unistate("st_light_pos", Vector3(0, 10, 0));
	set_unistate("st_light_color", Vector3(0.8, 0.8, 1.0));

	// ---- generate the tree ----
	rng_srand(234925);
	xtree = new PineTree;
	if(!xtree->generate(TREE_SIZE, opt.nbranches[0], opt.nbranches[1], opt.nbranches[2],
				opt.nbranches[3], opt.nbranches[4], 0)) {
		delete xtree;
		fatal_log("failed to generate christmas tree...\n");
		return false;
	}
	xtree->generate_lights(75, xtree->gen_params.lt_turns);

	// ---- environment ----
	floor_obj = new Object;
	floor_obj->set_mesh(new Mesh);
	//gen_plane(floor_obj->get_mesh(), 10, 10, 4, 4);
	gen_heightmap(floor_obj->get_mesh(), 10, 10, 5, 5, ground_heightmap);
	floor_obj->set_rotation(Quaternion(Vector3(1, 0, 0), -DEG_TO_RAD(90)));

	if(!(ground_sdr = get_sdrprog("default.v.glsl", "ground.p.glsl"))) {
		return false;
	}

	static const char *gndtex_name[] = {"grass02.jpg", "dirt01.jpg", 0};
	static const char *uni_name[] = {"tex_grass", "tex_dirt", 0};
	for(int i=0; gndtex_name[i]; i++) {
		if(!(floor_obj->material.tex[i] = texset.get(gndtex_name[i]))) {
			return false;
		}
		ground_sdr->set_uniform(uni_name[i], i);
	}

	floor_obj->material.specular = Vector3(1, 1, 1);
	floor_obj->material.shininess = 40.0;

	// ---- snowballs ----
	if(!(snowball_psdr = psdrset.get("snowball.p.glsl"))) {
		return false;
	}
	if(!(snowball_tex = texset.get("dirt01.jpg"))) {
		return false;
	}

	if(!psattr_snow_trail.load("snow_trail.psys")) {
		return false;
	}
	psattr_snow_trail.set_shader(snowball_psdr);

	if(!psattr_snow_hit.load("snow_hit.psys")) {
		return false;
	}

	ps_snowing = new ParticleSystem;
	if(!ps_snowing->load("snowfall.psys")) {
		return false;
	}
	ps_snowing->set_position(Vector3(-2, 7, 0));
	psman.add(ps_snowing);

	sball_mesh = new Mesh;
	gen_sphere(sball_mesh, 1.0, 8, 4);

	// ---- debug raytracer ----
	rt_scene_tree(xtree);
	dbg_rt_tex = new Texture2D;

	// ---- music ----
	music = new OggVorbisStream;

	if(!music->open("game_music.ogg")) {
		fatal_log("failed to load music\n");
		delete music;
		return false;
	}

	if(opt.music) {
		music->play(AUDIO_PLAYMODE_LOOP);
	}

	return true;
}

void GameScreen::cleanup()
{
	if(music) {
		music->stop();
		delete music;
	}

	delete xtree;
}

/*
static bool ispred(PineTreeBranch *branch, void *cls)
{
	// allow only intersections with branches which are children of the root
	return branch->parent && !branch->parent->parent;
}
*/

void GameScreen::update(unsigned long tmsec)
{
	static unsigned long time_paused;
	unsigned long dt_msec = tmsec - last_upd_time;
	float dt = dt_msec / 1000.0f;
	last_upd_time = tmsec;

	if(pause_updates) {
		time_paused += dt_msec;
		return;
	}

	// update the snowballs
	std::list<Snowball>::iterator it = snowballs.begin();
	while(it != snowballs.end()) {
		// if it's a new snowball, create its trail particle system too
		/*if(!it->psys_trail) {
			ParticleSystem *psys = new ParticleSystem;	// ownership goes to psman
			psys->set_attr(psattr_snow_trail);
			psman.add(psys);

			it->psys_trail = psys;
		}*/

		Vector3 newpos = it->pos + it->vel * dt;
		it->vel = it->vel * 0.99 + Vector3(0, -9.0, 0) * dt;

		// discard snowballs when they cross the ground plane
		if(newpos.y < 0.0) {
			if(it->psys_trail) {
				it->psys_trail->stop();	// will be deleted by the psys manager when its last particle is dead
			}
			it = snowballs.erase(it);
			debug_log("crossed the ground plane\n");
			continue;
		}

		Ray ray;
		ray.origin = it->pos;
		ray.dir = newpos - ray.origin;

		it->pos = newpos;
		it->life += dt;
		if(it->psys_trail) {
			it->psys_trail->set_position(newpos);
		}

		HitPoint pt;
		//if(xtree->sweep_intersect(ray, it->size, &pt) && pt.dist <= 1.0) {
		if(xtree->intersect(ray, &pt) && pt.dist <= 1.0) {
			dbg_last_sball_hit.center = pt.pos;
			dbg_last_sball_hit.radius = it->size * SNOWBALL_HIT_SCALE;

			xtree->extinguish(pt.pos, it->size * SNOWBALL_HIT_SCALE);
			it = snowballs.erase(it);

			ParticleSystem *psys = new ParticleSystem;
			psys->set_start_time(tmsec - time_paused);
			psys->set_attr(psattr_snow_hit);
			psys->set_position(pt.pos);
			psman.add(psys);	// psman has ownership
		} else {
			++it;
		}
	}

	psman.update(tmsec - time_paused);
	xtree->update(tmsec - time_paused);

	float life = xtree->life();

	music->set_pitch(life * 0.4 + 0.6);
	//music->set_volume(life);
}

void GameScreen::display() const
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	draw_backdrop(bg_horizon_color, bg_zenith_color);

	Matrix4x4 matrix;
	matrix.translate(Vector3(0, 0, -cam_dist));
	matrix.rotate(Vector3(1, 0, 0), DEG_TO_RAD(cam_phi));
	matrix.rotate(Vector3(0, 1, 0), DEG_TO_RAD(cam_theta));
	matrix.translate(-cam_pos);
	set_view_matrix(matrix);

	if(dbg_rt_pending) {
		rt_view_matrix(matrix);
		rt_render();
		Image *img = rt_get_image();
		dbg_rt_tex->set_image(*img);
		dbg_show_rt = 1;
		dbg_rt_pending = 0;
	}

	set_world_matrix(Matrix4x4::identity);

	bind_shader(ground_sdr);
	floor_obj->draw();

	bind_shader(defsdr);
	xtree->draw();
	//xtree->draw_bounds();

	// draw blended stuff
	draw_snowballs();
	xtree->draw_lights();
	xtree->draw_particles();
	psman.draw();

	bind_shader(0);
	set_world_matrix(Matrix4x4::identity);
	setup_gl_matrices();

#ifdef USE_OLDGL
	glBegin(GL_LINES);
	glColor3f(0, 1, 0);
	glVertex3f(dbg_ray.origin.x, dbg_ray.origin.y, dbg_ray.origin.z);
	Vector3 end = dbg_ray.origin + dbg_ray.dir * 1.0;
	glVertex3f(end.x, end.y, end.z);
	glEnd();

	if(dbg_hitpt.obj) {
		glPointSize(5.0);
		glBegin(GL_POINTS);
		glColor3f(1, 0, 1);
		glVertex3f(dbg_hitpt.pos.x, dbg_hitpt.pos.y, dbg_hitpt.pos.z);
		glEnd();
		glPointSize(1.0);
	}

	if(dbg_show_last_sball_hit) {
		static Mesh *mesh;
		if(!mesh) {
			mesh = new Mesh;
			gen_sphere(mesh, 1.0, 16, 8);
		}

		glPushAttrib(GL_ENABLE_BIT);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		glColor4f(0.2, 0.75, 1.0, 0.25);

		glMatrixMode(GL_MODELVIEW);
		glTranslatef(dbg_last_sball_hit.center.x, dbg_last_sball_hit.center.y, dbg_last_sball_hit.center.z);
		glScalef(dbg_last_sball_hit.radius, dbg_last_sball_hit.radius, dbg_last_sball_hit.radius);
		mesh->draw();

		glPopAttrib();
	}
#endif	/* USE_OLDGL */

	glDisable(GL_DEPTH_TEST);

#ifdef USE_OLDGL
	/* show raytracing output if requested */
	if(dbg_show_rt) {
		glPushAttrib(GL_ENABLE_BIT);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		//glTranslatef(-0.6, 0.6, 0);
		//glScalef(0.4, 0.4, 0.4);

		glEnable(GL_TEXTURE_2D);
		set_texture(dbg_rt_tex, 0);

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		glBegin(GL_QUADS);
		glColor4f(1, 1, 1, 0.5);
		glTexCoord2f(0, 1); glVertex2f(-1, -1);
		glTexCoord2f(1, 1); glVertex2f(1, -1);
		glTexCoord2f(1, 0); glVertex2f(1, 1);
		glTexCoord2f(0, 0); glVertex2f(-1, 1);
		glEnd();
		glDisable(GL_BLEND);

		glPopAttrib();
	}
#endif	// USE_OLDGL

	/* print score */

	dtx_use_font(font, FONT_SIZE);
	dtx_vertex_attribs(font_sdr_vattr, font_sdr_tcattr);

	char score_text[256];
	sprintf(score_text, "Life: %3d %%", (int)(xtree->life() * 100.0f));
	float score_width = dtx_string_width(score_text);

	Matrix4x4 score_xform;
	score_xform.set_translation(Vector3(get_screen_width() - score_width - 5, get_screen_height() - 20, 0));
	set_world_matrix(score_xform);
	set_unistate("st_color", Vector4(1, 1, 1, 1));
	font_sdr->bind();

	dtx_string(score_text);
	dtx_flush();

	glEnable(GL_DEPTH_TEST);
}

static void draw_snowballs()
{
	bind_shader(defsdr);
	//point_pixel_shader(snowball_psdr);
	snowball_tex->bind();

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	//float color[] = {1.0, 1.0, 1.0, 1.0};

	//begin_points();
	std::list<Snowball>::const_iterator it = snowballs.begin();
	while(it != snowballs.end()) {
		//color[3] = std::min(it->life * 16.0f, 1.0f);
		//draw_point(&it->pos[0], color, it->size);

		Matrix4x4 xform;
		xform.set_translation(it->pos);
		xform.scale(Vector3(it->size, it->size, it->size));
		set_world_matrix(xform);
		setup_unistate();

		sball_mesh->draw();
		++it;
	}
	//end_points();

	glDisable(GL_BLEND);
	point_pixel_shader(0);
}

void GameScreen::reshape(int x, int y)
{
	glViewport(0, 0, x, y);

	win_width = x;
	win_height = y;

	Matrix4x4 proj;
	proj.set_perspective(M_PI / 4.0, (float)x / (float)y, 0.5, 500.0);
	set_projection_matrix(proj);

	rt_resize(x / 4, y / 4);
}

void GameScreen::keyboard(int key, bool pressed)
{
	if(pressed) {
		switch(key) {
		case '\b':
			dbg_switch = !dbg_switch;
			break;

		case '=':
			dbg_iter++;
			break;

		case '-':
			dbg_iter--;
			break;

		case ' ':
			pause_updates = !pause_updates;
			break;

		case 'r':
			dbg_rt_pending = 1;
			break;

		case 'R':
			dbg_show_rt = !dbg_show_rt;
			if(!dbg_show_rt) {
				dbg_rt_pending = 0;
			}
			break;

		case 'b':
			dbg_show_last_sball_hit = !dbg_show_last_sball_hit;
			break;
		}
	}
}

/* distance threshold of mouse movement in pixels, above which it's considered
 * a drag, as opposed to a click. */
#define CLICK_THRES	std::max(5.0f, win_height * 0.025f)

void GameScreen::button(int bn, bool pressed, int x, int y)
{
	static bool firestarter;

	prev_x = x;
	prev_y = y;
	bnstate[bn] = pressed;

	if(pressed) {
		press_x[bn] = x;
		press_y[bn] = y;
		abs_dx = abs_dy = 0;
	} else {
		if(abs_dx + abs_dy < CLICK_THRES) {
			// XXX debug stuff
			if(x < 70 && y < 100) {
				firestarter = !firestarter;
				return;
			}

			if(bn == 0 && !firestarter) {
				throw_snow(x, y, snowball_size);
			}
			if(bn == 1) {
				dbg_ray = calc_pick_ray(x, y);
				if(!xtree->intersect(dbg_ray, &dbg_hitpt)) {
					dbg_hitpt.obj = 0;
				}
			}
			if(bn == 2 || (bn == 0 && firestarter)) {
				// XXX for testing, ignite on right-click
				Ray ray = calc_pick_ray(x, y);
				HitPoint hit;
				if(xtree->intersect(ray, &hit)) {
					xtree->ignite(hit.pos, 0.2);
				}
			}
		}
	}
}

void GameScreen::motion(int x, int y, bool pressed)
{
	int dx = x - prev_x;
	int dy = y - prev_y;
	prev_x = x;
	prev_y = y;

	abs_dx += abs(dx);
	abs_dy += abs(dy);

	if(!dx && !dy) return;

	if(bnstate[0]) {
		cam_theta += dx * 0.5;
		if(cam_phi > 90) cam_phi = 90;

		if(get_key_state(KEY_SHIFT)) {
			cam_phi += dy * 0.5;
			if(cam_phi < -90) cam_phi = -90;
		}
	}

	if(bnstate[1]) {
		float theta_rad = DEG_TO_RAD(cam_theta);

		cam_pos.x += (cos(theta_rad) * dx - sin(theta_rad) * dy) * 0.05;
		cam_pos.z += (sin(theta_rad) * dx + cos(theta_rad) * dy) * 0.05;
	}

	if(bnstate[2]) {
		cam_dist += dy * 0.1;

		if(cam_dist < 0) cam_dist = 0;
	}
}

long GameScreen::redisplay_interval() const
{
	return 1;
}

static void throw_snow(int x, int y, float sbsize)
{
	if(sbsize < SNOWBALL_SIZE_MIN) sbsize = SNOWBALL_SIZE_MIN;
	if(sbsize > SNOWBALL_SIZE_MAX) sbsize = SNOWBALL_SIZE_MAX;

	last_throw_time = last_upd_time;

	Ray sbray = calc_pick_ray(x, y);

	/*info_log("throwing snow with sbsize: %g, pos: %g %g %g -> dir: %g %g %g\n", sbsize,
			sbray.origin.x, sbray.origin.y, sbray.origin.z, sbray.dir.x, sbray.dir.y, sbray.dir.z);
	*/

	Snowball ball;
	ball.size = sbsize;
	ball.pos = sbray.origin;
	ball.vel = (sbray.dir.normalized() + Vector3(0, 0.1, 0)) * 18.0;
	snowballs.push_back(ball);
}

#undef USE_GLU

static Ray calc_pick_ray(int x, int y)
{
	Matrix4x4 vmat = get_view_matrix();
	Matrix4x4 pmat = get_projection_matrix();

#ifndef USE_GLU
	float px = 2.0 * (float)x / win_width - 1.0;
	float py = 1.0 - (2.0 * (float)y / win_height);

	Matrix4x4 mvp_inv = (pmat * vmat).inverse();

	Vector4 vn = Vector4(px, py, -1, 1).transformed(mvp_inv);
	Vector4 vf = Vector4(px, py, 1, 1).transformed(mvp_inv);

	vn /= vn.w;
	vf /= vf.w;
	vn.w = vf.w = 1.0;
#else
	double vm[16], pm[16];
	int vp[4];

	glGetIntegerv(GL_VIEWPORT, vp);

	for(int i=0; i<4; i++) {
		for(int j=0; j<4; j++) {
			vm[j * 4 + i] = vmat[i][j];
			pm[j * 4 + i] = pmat[i][j];
		}
	}

	double res[3];
	gluUnProject(x, win_height - y, 0, vm, pm, vp, res, res + 1, res + 2);
	Vector3 vn = Vector3(res[0], res[1], res[2]);
	gluUnProject(x, win_height - y, 1, vm, pm, vp, res, res + 1, res + 2);
	Vector3 vf = Vector3(res[0], res[1], res[2]);
#endif

	return Ray(vn, vf - vn);
}

float ground_heightmap(float u, float v, void *cls)
{
	float x = u * 2.0 - 1.0;
	float y = v * 2.0 - 1.0;


	float rad = sqrt(x * x + y * y) / 3.0;

	u /= 3.0;
	v /= 3.0;
	return fbm2(u * 5.0, v * 5.0, 2) * 0.3 - 1.0 * rad + 0.1;
}
