#ifdef USE_GLUT

#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include "opengl.h"
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include "options.h"
#include "game.h"
#include "input.h"
#include "screen.h"
#include "logger.h"

/* called by any other part of the game when a redisplay is needed
 * (mostly useful for tools which do not redisplay continuously)
 */
void request_redisplay();

void swap_buffers();

static void display(void);
static void reshape(int x, int y);
static void keydown(unsigned char key, int x, int y);
static void keyup(unsigned char key, int x, int y);
static void skeydown(int key, int x, int y);
static void skeyup(int key, int x, int y);
static void mouse(int bn, int state, int x, int y);
static void motion(int x, int y);
static void passive_motion(int x, int y);

static long redisp_interval;

int main(int argc, char **argv)
{
	glutInit(&argc, argv);

	load_config("data/game.cfg");
	if(!parse_args(argc, argv)) {
		return 1;
	}

	glutInitWindowSize(opt.width, opt.height);
	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE | GLUT_MULTISAMPLE);
	glutCreateWindow("Xmass inferno");

	if(opt.fullscreen) {
		glutFullScreen();
	}

	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keydown);
	glutKeyboardUpFunc(keyup);
	glutSpecialFunc(skeydown);
	glutSpecialUpFunc(skeyup);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutPassiveMotionFunc(passive_motion);

	if(!game_init()) {
		return 1;
	}
	atexit(game_cleanup);

	glutMainLoop();
	return 0;
}

void request_redisplay()
{
	glutPostRedisplay();
}

void swap_buffers()
{
	glUseProgram(0);
	glutSwapBuffers();
}

static void display()
{
	long ival = current_screen()->redisplay_interval();
	if(ival && !redisp_interval) {
		debug_log("switching to auto-redisplay (interval: %ld)\n", redisp_interval);
		glutIdleFunc(request_redisplay);
	}
	if(!ival && redisp_interval) {
		debug_log("switching to manual redisplay\n");
		glutIdleFunc(0);
	}
	redisp_interval = ival;

	game_display();
	// TODO implement framerate limiter
}


static void reshape(int x, int y)
{
	game_reshape(x, y);
}

/* GLUT doesn't have any callbacks for modifier keys, so we fake it by calling
 * this function all over the place, to generate fake events whenever a modifier
 * key is detected to have changed state.
 * The limitation is that this will only detect modkey changes whenever a regular
 * key is pressed or released, or when a mouse button is pressed or released.
 */
static void handle_modkeys()
{
	static int modstate;

	int mod = glutGetModifiers();
	if(mod != modstate) {
		int diff = mod ^ modstate;

		if(diff & GLUT_ACTIVE_SHIFT) {
			bool pressed = mod & GLUT_ACTIVE_SHIFT;
			game_key(KEY_LSHIFT, pressed);
		}
		if(diff & GLUT_ACTIVE_CTRL) {
			bool pressed = mod & GLUT_ACTIVE_CTRL;
			game_key(KEY_LCTRL, pressed);
		}
		if(diff & GLUT_ACTIVE_ALT) {
			bool pressed = mod & GLUT_ACTIVE_ALT;
			game_key(KEY_LALT, pressed);
		}
	}
	modstate = mod;
}

static int translate_special_key(int key)
{
	if(key >= GLUT_KEY_F1 && key <= GLUT_KEY_F12) {
		return (key - GLUT_KEY_F1) + KEY_F1;
	}
	if(key >= GLUT_KEY_LEFT && key <= GLUT_KEY_PAGE_DOWN) {
		return (key - GLUT_KEY_LEFT) + KEY_LEFT;
	}

	switch(key) {
	case GLUT_KEY_HOME:
		return KEY_HOME;
	case GLUT_KEY_END:
		return KEY_END;
	case GLUT_KEY_INSERT:
		return KEY_INSERT;

	default:
		break;
	}
	return 0;
}

static void keydown(unsigned char c, int x, int y)
{
	int key = c;
	if(key == 127) {
		key = KEY_DELETE;
	}

	handle_modkeys();
	game_key(key, true);
}

static void keyup(unsigned char c, int x, int y)
{
	int key = c;
	if(key == 127) {
		key = KEY_DELETE;
	}

	handle_modkeys();
	game_key(key, false);
}

static void skeydown(int key, int x, int y)
{
	handle_modkeys();

	if((key = translate_special_key(key)) > 0) {
		game_key(key, true);
	}
}

static void skeyup(int key, int x, int y)
{
	handle_modkeys();

	if((key = translate_special_key(key)) > 0) {
		game_key(key, false);
	}
}

static void mouse(int bn, int state, int x, int y)
{
	handle_modkeys();

	int bnidx = bn - GLUT_LEFT_BUTTON;
	bool bnstate = state == GLUT_DOWN;

	game_mbutton(0, bnidx, bnstate, x, y);
}

static void motion(int x, int y)
{
	game_mmotion(0, x, y, true);
}

static void passive_motion(int x, int y)
{
	game_mmotion(0, x, y, false);
}
#endif	// USE_GLUT
