#ifdef USE_SDL

#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
#include "opengl.h"
#include <SDL2/SDL.h>

#include "options.h"
#include "game.h"
#include "input.h"
#include "screen.h"
#include "logger.h"
#include "shader.h"

/* called by any other part of the game when a redisplay is needed
 * (mostly useful for tools which do not redisplay continuously)
 */
void request_redisplay();

void swap_buffers();

static void cleanup(void);
static void display(void);
static void reshape(int x, int y);
static bool handle_event(SDL_Event *ev);

static SDL_Window *win;
static SDL_GLContext ctx;
static bool redisplay_pending;
static long redisp_interval;

int main(int argc, char **argv)
{
	SDL_Init(SDL_INIT_VIDEO);

	load_config("data/game.cfg");
	if(!parse_args(argc, argv)) {
		return false;
	}

	int undef = SDL_WINDOWPOS_UNDEFINED;
	unsigned int flags = SDL_WINDOW_OPENGL | SDL_WINDOW_RESIZABLE;
	if(opt.fullscreen) {
		flags |= SDL_WINDOW_FULLSCREEN_DESKTOP;
	}
	if(!(win = SDL_CreateWindow("xmas inferno", undef, undef, opt.width, opt.height, flags))) {
		fprintf(stderr, "failed to create window\n");
		return -1;
	}
	if(!(ctx = SDL_GL_CreateContext(win))) {
		fprintf(stderr, "failed to create OpenGL context\n");
		return -1;
	}

	if(!game_init()) {
		return 1;
	}
	atexit(cleanup);

	int width, height;
	SDL_GetWindowSize(win, &width, &height);
	game_reshape(width, height);

	for(;;) {
		SDL_Event ev;
		while(SDL_PollEvent(&ev)) {
			if(!handle_event(&ev)) {
				goto done;
			}
		}

		redisplay_pending = false;
		display();
	}

done:
	cleanup();
	return 0;
}

void cleanup()
{
	game_cleanup();
	SDL_Quit();
}

void request_redisplay()
{
	redisplay_pending = true;
}

void swap_buffers()
{
	bind_shader(0);
	SDL_GL_SwapWindow(win);
}

static void display()
{
	long ival = current_screen()->redisplay_interval();
	if(ival && !redisp_interval) {
		debug_log("switching to auto-redisplay (interval: %ld)\n", redisp_interval);
		redisplay_pending = true;
	}
	if(!ival && redisp_interval) {
		debug_log("switching to manual redisplay\n");
	}
	redisp_interval = ival;

	game_display();
	// TODO implement framerate limiter
}


static void reshape(int x, int y)
{
	game_reshape(x, y);
}

bool handle_event(SDL_Event *ev)
{
	switch(ev->type) {
	case SDL_QUIT:
		return false;

	case SDL_KEYDOWN:
	case SDL_KEYUP:
		{
			int key = ev->key.keysym.sym & 0xffff;
			switch(key) {
			case SDLK_DELETE:
				key = KEY_DELETE;
				break;
			case SDLK_LSHIFT:
				key = KEY_LSHIFT;
				break;
			case SDLK_RSHIFT:
				key = KEY_RSHIFT;
				break;
			default:
				break;
			}

			game_key(key, ev->key.state == SDL_PRESSED);
		}
		break;

	case SDL_MOUSEBUTTONDOWN:
	case SDL_MOUSEBUTTONUP:
		{
			int bn = ev->button.button - SDL_BUTTON_LEFT;
			bool pressed = ev->button.state == SDL_PRESSED;
			int x = ev->button.x;
			int y = ev->button.y;

			current_screen()->button(bn, pressed, x, y);
			set_mouse_state(x, y, bn, pressed);	// this MUST be after the event dispatch
		}
		break;

	case SDL_MOUSEMOTION:
		{
			int x = ev->motion.x;
			int y = ev->motion.y;
			bool drag = ev->motion.state != 0;

			set_mouse_state(x, y);
			current_screen()->motion(x, y, drag);
		}
		break;

	case SDL_WINDOWEVENT:
		if(ev->window.event == SDL_WINDOWEVENT_RESIZED) {
			reshape(ev->window.data1, ev->window.data2);
		}
		break;

	default:
		break;
	}
	return true;
}

#endif	// USE_SDL
