#include <float.h>
#include <assert.h>
#include <algorithm>
#include "xtree.h"
#include "object.h"
#include "sdrman.h"
#include "texman.h"
#include "unistate.h"
#include "game.h"
#include "options.h"
#include "logger.h"


extern bool dbg_switch;
extern int dbg_iter;

PineTreeBranch::PineTreeBranch()
{
	spawn_pt = 0.0;
	amb_occ[0] = amb_occ[1] = 1.0;

	burned = 0.0;
	onfire = false;
	burn_start_msec = 0;
	size = 1.0;

	parent = 0;
	tree = 0;
}

bool PineTreeBranch::get_burn_fronts(float *b0, float *b1)
{
	if(!onfire || burned >= 1.0) {
		*b0 = 0;
		*b1 = 1;
		return false;
	}

	float max_dist = fire_src > 0.5 ? fire_src : (1.0 - fire_src);
	float dist = burned * max_dist;
	*b0 = std::max(fire_src - dist, 0.0f);
	*b1 = std::min(fire_src + dist, 1.0f);
	return true;
}

void PineTreeBranch::calc_bounds()
{
	Vector3 branch_dir = end[1] - end[0];
	float branch_len = branch_dir.length();
	branch_dir /= branch_len;

	float x0 = branch_size(&tree->gen_params, branch_len, 1.0);
	float x1 = branch_size(&tree->gen_params, branch_len, 0.0);
	float dx = x1 - x0;
	assert(dx >= 0.0);

	float tan_theta = dx / branch_len;
	float theta = atan(tan_theta);
	float boffs = branch_len * tree->gen_params.scale_offs / tan_theta;

	bcone.apex = end[1] + branch_dir * boffs;
	bcone.dir = -branch_dir;
	bcone.begin = boffs;
	bcone.end = size + boffs;
	bcone.set_angle(theta);
}


bool PineTreeBranch::intersect(const Ray &ray, HitPoint *pt) const
{
	return intersect(ray, pt, 0);
}

bool PineTreeBranch::intersect(const Ray &ray, HitPoint *pt, bool (*pred)(PineTreeBranch*, void*), void *cls) const
{
	HitPoint dummy;
	if(!pt) pt = &dummy;

	if(!bcone.contains(ray.origin) && (!bcone.intersect(ray, pt) || pt->dist > 1.0f)) {
		return false;
	}

	if(children.empty()) {
		return true;
	}

	pt->dist = FLT_MAX;
	pt->obj = 0;

	for(size_t i=0; i<children.size(); i++) {
		HitPoint hit;
		if(children[i]->intersect(ray, &hit, pred, cls) &&
				hit.dist < pt->dist && (!pred || pred(children[i], cls))) {
			*pt = hit;
		}
	}

	return pt->obj != 0;
}

#define SWEEP_STEPS	3
bool PineTreeBranch::sweep_intersect(const Ray &ray, float rad, HitPoint *pt) const
{
	Ray bray;	// branch ray
	bray.origin = end[0];
	bray.dir = end[1] - end[0];

	HitPoint hit;
	hit.obj = 0;
	Sphere sph;
	sph.radius = rad;
	for(int i=0; i<SWEEP_STEPS; i++) {
		float t = (float)i / (float)(SWEEP_STEPS - 1);
		sph.center = ray.origin + ray.dir * t;

		if(sph.intersect(bray, &hit) && hit.dist <= 1.0f) {
			break;
		}
	}

	if(children.empty()) {
		if(pt) *pt = hit;
		return true;
	}

	HitPoint dummy;
	if(!pt) pt = &dummy;
	*pt = hit;

	for(size_t i=0; i<children.size(); i++) {
		HitPoint chit;
		if(children[i]->sweep_intersect(ray, rad, &chit) && chit.dist < pt->dist) {
			*pt = chit;
		}
	}

	return pt->obj != 0;
}

float PineTreeBranch::distance(const Vector3 &pt) const
{
	if(children.empty()) {
		return bcone.distance(pt);
	}

	float best = FLT_MAX;
	for(size_t i=0; i<children.size(); i++) {
		float d = children[i]->distance(pt);
		if(fabs(d) < fabs(best)) {
			best = d;
		}
	}
	return best;
}


float PineTree::burn_rate = 0.1;	/* world-space units per second */

PineTree::PineTree()
{
	bark_sdr = needle_sdr = 0;
	root = 0;
	root_branch = 0;
	prev_upd = 0;
	size = 1.0;

	init_gen();

	bark_sdr = needle_sdr = 0;

	needles_lbr_meshes = 0;
}

PineTree::~PineTree()
{
	clear();
}

bool PineTree::init()
{
	if(!(bark_sdr = get_sdrprog("default.v.glsl", "diffuse_tex.p.glsl"))) {
		return false;
	}
	if(!(needle_sdr = get_sdrprog("needles.v.glsl", "needles.p.glsl"))) {
		return false;
	}
	if((uloc_needle_fire_interval = needle_sdr->get_uniform_location("fire_interval")) == -1) {
		debug_log("needle shader doesn't have the fire interval uniform\n");
	}

	if(!(bark_tex = texset.get("tree_bark1.jpg"))) {
		return false;
	}

	if(!(needle_tex = texset.get("pine_needles.png"))) {
		return false;
	}
	needle_tex->set_wrapping(GL_REPEAT, GL_CLAMP_TO_EDGE);

	if(!(light_tex = texset.get("glow1.png"))) {
		return false;
	}

	ps_fire.set_parent_tree(this);
	if(!ps_fire.load("fire.psys")) {
		return false;
	}
	ps_fire.set_position(Vector3(0, 2, 0));

	if(opt.smoke) {
		ps_smoke.set_parent_tree(this);
		if(!ps_smoke.load("smoke.psys")) {
			return false;
		}
		ps_fire.add_child(&ps_smoke);
		ps_smoke.set_position(Vector3(0, 0, 0));
	}

	return true;
}

void PineTree::clear()
{
	for(size_t i=0; i<branches.size(); i++) {
		delete branches[i];
	}
	branches.clear();

	inner_branches.clear();
	leaf_branches.clear();
	burning_branches.clear();

	for(size_t i=0; i<leaves.size(); i++) {
		delete leaves[i];
	}
	leaves.clear();

	root_branch = 0;
	root = 0;
	prev_upd = 0;
	size = 1.0;

	Scene::clear();

	lights.clear();

	delete [] needles_lbr_meshes;
	needles_lbr_meshes = 0;
}


float PineTree::life() const
{
	float life_acc = 0.0f;
	int num_leaf_br = (int)leaf_branches.size();

	for(int i=0; i<num_leaf_br; i++) {
		life_acc += 1.0 - leaf_branches[i]->burned;
	}

	return life_acc / (float)num_leaf_br;
}

float PineTree::radius(float y) const
{
	if(!root) return -1.0;

	Vector3 pos = root->get_position();
	float t = (y - pos.y) / size;

	if(t < 0.0 || t > 1.0) return -1.0;

	return branch_size(&gen_params, size, t);
}

bool PineTree::coarse_intersect(const Ray &ray, HitPoint *pt) const
{
	if(!root_branch) {
		return false;
	}
	return root_branch->bcone.intersect(ray, pt);
}

bool PineTree::intersect(const Ray &ray, HitPoint *pt) const
{
	return intersect(ray, pt, 0, 0);
}

bool PineTree::intersect(const Ray &ray, HitPoint *pt,
		bool (*pred)(PineTreeBranch*, void*), void *cls) const
{
	if(!root_branch) {
		return false;
	}
	return root_branch->intersect(ray, pt, pred, cls);
}

bool PineTree::sweep_intersect(const Ray &ray, float rad, HitPoint *pt) const
{
	if(!root_branch) {
		return false;
	}
	return root_branch->sweep_intersect(ray, rad, pt);
}

float PineTree::distance(const Vector3 &pt) const
{
	if(!root_branch) {
		return FLT_MAX;
	}
	return root_branch->distance(pt);
}

void PineTree::ignite(PineTreeBranch *br, float pos)
{
	if(!br) return;

	if(!br->onfire && br->burned < 1.0) {
		br->onfire = true;
		br->burn_start_msec = -1;	/* next update will set the start time */
		br->fire_src = pos;
		burning_branches.push_back(br);
	}
}

void PineTree::extinguish(PineTreeBranch *br)
{
	if(!br) return;

	br->onfire = false;

	std::list<PineTreeBranch*>::iterator it = burning_branches.begin();
	while(it != burning_branches.end()) {
		if(*it == br) {
			burning_branches.erase(it);
			return;
		}
		++it;
	}
}

void PineTree::ignite(const Vector3 &p, float rad)
{
	// TODO accelerate branch searching
	for(size_t i=0; i<branches.size(); i++) {
		PineTreeBranch *br = branches[i];

		// XXX only ignite leaf branches?
		// otherwise at the very least we'd have to skip the root cone...
		if(!br->children.empty()) {
			continue;
		}

		if(br->bcone.distance(p) <= rad) {
			// find approximate position on the branch
			Vector3 bdir = (br->end[1] - br->end[0]) / br->size;
			float bpos = dot_product(bdir, p - br->end[0]) / br->size;
			debug_log("igniting branch: %p\n", (void*)br);
			ignite(br, std::max(0.0f, std::min(bpos, 1.0f)));
		}
	}
}

/* we want to be able to extinguish fires on internal branches, otherwise
 * we can't stop the spread once it leaves the leaves (sic). However, we can't
 * just use the inner branches' bounding volumes for this or we'll always hit
 * only internal branches.
 * So this is a hybrid approach which checks bounding volumes for leaf nodes
 * but ray(branch)-sphere tests for internal ones.
 */
void PineTree::extinguish(const Vector3 &p, float rad)
{
	int num = 0;

	// same as above
	Sphere sph = Sphere(p, rad);

	std::list<PineTreeBranch*>::iterator it = burning_branches.begin();
	while(it != burning_branches.end()) {
		PineTreeBranch *br = *it++;

		if(!br->children.empty()) {
			// this is an internal branch: use ray-testing to extinguish
			Ray ray = Ray(br->end[0], br->end[1] - br->end[0]);

			HitPoint hit;
			if(sph.contains(br->end[0]) || (sph.intersect(ray, &hit) && hit.dist <= 1.0)) {
				//debug_log("extinguishing inner branch: %p\n", (void*)br);
				extinguish(br);
				++num;
			}

		} else {
			// this is a leaf branch: use sphere-cone collision to extinguish
			if(br->bcone.distance(p) <= rad) {
				//debug_log("extinguishing leaf branch: %p\n", (void*)br);
				extinguish(br);
				++num;
			}
		}
	}

	debug_log("extinguished %d branches, %d remaining\n", num, (int)burning_branches.size());
}

void PineTree::update(long msec)
{
	float dt = (msec - prev_upd) / 1000.0f;
	prev_upd = msec;

	//debug_log("# burning branches: %d\n", (int)burning_branches.size());
	std::list<PineTreeBranch*>::iterator it = burning_branches.begin();
	while(it != burning_branches.end()) {
		PineTreeBranch *br = *it;
		assert(br->onfire);
		if(br->onfire && br->burned < 1.0) {
			++it;
		} else {
			br->onfire = false;
			it = burning_branches.erase(it);
			continue;
		}

		if(br->burn_start_msec == -1) {
			br->burn_start_msec = msec;	// just starting to burn
		}

		float adv = burn_rate * dt / br->size;
		br->burned = std::min(br->burned + adv, 1.0f);

		float b0, b1;
		br->get_burn_fronts(&b0, &b1);

		// if the front has reached the start of the branch, ignite the parent
		if(b0 < 1e-5) {
			ignite(br->parent, br->spawn_pt);
		}

		// ignite any children in the burn interval
		int num_children = (int)br->children.size();
		for(int i=0; i<num_children; i++) {
			float ct = br->children[i]->spawn_pt;
			if(ct >= b0 && ct < b1) {
				ignite(br->children[i], 0.0);
			}
		}
	}

	// update particle systems
	ps_fire.update(msec);
	if(opt.smoke) {
		ps_smoke.update(msec);
	}
}

float branch_size(const PineTreeParams *param, float size, float t)
{
	return size * (param->scale_offs + param->scale * (1.0 - t));
}
