#ifndef XTREE_H_
#define XTREE_H_

#include <vector>
#include <list>
#include "vmath/vmath.h"
#include "scene.h"
#include "shader.h"
#include "geom.h"
#include "texture.h"
#include "image.h"
#include "psys_fire.h"

class PineTree;
class PineTreeBranch;


struct TreeLight {
	bool alive;
	Vector3 pos;
	float intensity;
	Vector3 color;
	float blink_speed, blink_offs;
	float size;
};

struct PineTreeParams {
	float scale;		// scale of branches relative to their parent
	float rad_scale;	// scale of branch radii relative to their parent
	float scale_offs;	// added to the branch size (also relative)
	float angle;		// outward angle of the branches in degrees (0 = parent)
	float spawn_min, spawn_max;	// spawn interval along the branch
	float spiral_turns;
	float distr_bias;
	// light parameters
	float lt_blink_speed;
	float lt_distr_bias;
	float lt_turns;
	float lt_spawn_min, lt_spawn_max;
	// pine needle parameters
	float needles_density;	// needle bunches per unit branch size
};

class PineTreeBranch {
public:
	float spawn_pt;	// spawn point (parametric) along the parent
	Vector3 end[2];	// world pos of branch ends
	float amb_occ[2];	// approximate ambient occlusion values for each end
	float size;		// (end[1] - end[0]).length()
	float burned;	// burn fraction, starts at 0.0
	bool onfire;
	long burn_start_msec;	// when did it start burning?
	float fire_src;	// starting position of the fire along the branch

	PineTree *tree;			// the tree we belong to
	PineTreeBranch *parent;	// parent branch
	std::vector<PineTreeBranch*> children;	// sub-branches

	Cone bcone;	// bounding cone (see calc_bounds())

	Matrix4x4 xform;	// not used much

	PineTreeBranch();

	bool get_burn_fronts(float *b0, float *b1);

	void calc_bounds();
	void draw_bounds() const;

	bool intersect(const Ray &ray, HitPoint *pt = 0) const;
	bool intersect(const Ray &ray, HitPoint *pt, bool (*pred)(PineTreeBranch*, void*), void *cls = 0) const;
	bool sweep_intersect(const Ray &ray, float rad, HitPoint *pt = 0) const;
	float distance(const Vector3 &pt) const;
};

class PineTree : public Scene {
private:
	long prev_upd;
	ShaderProg *bark_sdr, *needle_sdr;
	int uloc_needle_fire_interval;
	Texture *bark_tex, *needle_tex;
	Texture *light_tex;

	Mesh needles_mesh;
	// XXX optimization experiment
	Mesh *needles_lbr_meshes;

	FireParticleSystem ps_fire, ps_smoke;

	void init_gen();

public:
	float size;
	PineTreeParams gen_params;

	std::vector<Object*> leaves;
	Object *root;
	std::vector<PineTreeBranch*> inner_branches;
	std::vector<PineTreeBranch*> leaf_branches;
	std::vector<PineTreeBranch*> branches;
	std::list<PineTreeBranch*> burning_branches;
	PineTreeBranch *root_branch;

	std::vector<TreeLight> lights;

	static float burn_rate;

	PineTree();
	~PineTree();

	bool init();
	void clear();

	bool generate(float size, ...);
	bool generate_lights(int num, float turns);

	float life() const;

	// the radius of the tree at world-space height y, or -1.0 if below/above
	float radius(float y) const;

	bool coarse_intersect(const Ray &ray, HitPoint *pt = 0) const;
	bool intersect(const Ray &ray, HitPoint *pt = 0) const;
	bool intersect(const Ray &ray, HitPoint *pt, bool (*pred)(PineTreeBranch*, void*), void *cls = 0) const;
	bool sweep_intersect(const Ray &ray, float rad, HitPoint *pt = 0) const;
	float distance(const Vector3 &pt) const;

	void ignite(PineTreeBranch *br, float pos = 0.5);
	void extinguish(PineTreeBranch *br);

	void ignite(const Vector3 &p, float rad = 1.0);
	void extinguish(const Vector3 &p, float rad = 1.0);

	void update(long msec);
	void draw(long msec = 0) const;
	void draw_particles() const;
	void draw_needles() const;
	void draw_lights() const;
	void draw_bounds() const;
};

// helper functions
float branch_size(const PineTreeParams *param, float size, float t);

#endif	// XTREE_H_
