// the rendering part of the tree code
#include "xtree.h"
#include "unistate.h"
#include "points.h"
#include "options.h"

#define BRANCH_ALPHA	1.0
#define LEAVES_ALPHA	0.5
#define BVOL_ALPHA		0.5

void PineTreeBranch::draw_bounds() const
{
#ifdef USE_OLDGL
	float height = bcone.end;
	Vector3 vbeg = bcone.apex + bcone.dir * bcone.begin;
	Vector3 vend = bcone.apex + bcone.dir * bcone.end;

	glPointSize(3.0);
	glBegin(GL_POINTS);
	glColor4f(1, 0.3, 0.2, BVOL_ALPHA);
	glVertex3f(bcone.apex.x, bcone.apex.y, bcone.apex.z);
	glColor4f(0.8, 0.8, 0.1, BVOL_ALPHA);
	glVertex3f(vbeg.x, vbeg.y, vbeg.z);
	glVertex3f(vend.x, vend.y, vend.z);
	glEnd();


	Vector3 up = fabs(bcone.dir.y) < 0.99 ? Vector3(0, 1, 0) : Vector3(0, 0, 1);
	Vector3 right = cross_product(bcone.dir, up).normalized();

	// draw the bounding cone
	float slope = bcone.get_slope();

	int bcone_seg = 10, circle_seg = 16;
	glColor4f(0.2, 0.3, 0.8, BVOL_ALPHA);
	for(int i=0; i<bcone_seg; i++) {
		float t = (float)i / (float)(bcone_seg - 1);
		float rad = slope * t * height;

		Vector3 c = bcone.apex + bcone.dir * t * height;

		Matrix4x4 rot;

		glBegin(GL_LINE_LOOP);
		for(int j=0; j<circle_seg; j++) {
			float u = (float)j / (float)circle_seg;
			float theta = u * 2.0 * M_PI;

			rot.set_rotation(bcone.dir, theta);
			Vector3 v = c + right.transformed(rot) * rad;

			glVertex3f(v.x, v.y, v.z);
		}
		glEnd();
	}
#endif
}

void PineTree::draw(long msec) const
{
	bind_shader(bark_sdr);
	Scene::draw(msec);

	set_unistate("st_tex_matrix", Matrix4x4::identity);
	draw_needles();
}

void PineTree::draw_particles() const
{
	if(opt.smoke) {
		ps_smoke.draw();
	}
	ps_fire.draw();
}

void PineTree::draw_needles() const
{
	//PineTreeParams param = gen_params;
	//param.scale *= 0.5;

	if(needle_tex) {
		needle_tex->bind();
	}
	needle_sdr->bind();

	int nleafbr = leaf_branches.size();
	for(int i=0; i<nleafbr; i++) {
		PineTreeBranch *br = leaf_branches[i];

		float fire0, fire1;
		if(br->burned >= 1.0) {
			fire0 = 0;
			fire1 = 1;
		} else {
			if(!br->get_burn_fronts(&fire0, &fire1)) {
				fire0 = fire1 = 0;
			}
		}
		needle_sdr->set_uniform(uloc_needle_fire_interval, Vector2(fire0, fire1));
		needles_lbr_meshes[i].draw();
	}
}

void PineTree::draw_lights() const
{
	glDepthMask(0);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);

	light_tex->bind();
	begin_points();

	int num_lights = (int)lights.size();
	for(int i=0; i<num_lights; i++) {
		const TreeLight *lt = &lights[i];

		Vector3 col = lt->color * lt->intensity;
		point_color(col.x, col.y, col.z, 1.0);

		draw_point(&lt->pos.x, lt->size);
	}

	end_points();

	glDisable(GL_BLEND);
	glDepthMask(1);
}

void PineTree::draw_bounds() const
{
	glUseProgram(0);
	set_world_matrix(Matrix4x4::identity);
	setup_gl_matrices();

	for(size_t i=0; i<branches.size(); i++) {
		if(branches[i]->parent != root_branch) {
			continue;
		}
		branches[i]->draw_bounds();
	}
}
