// the procedural generation (init) part of xtree
#include <stdarg.h>
#include "xtree.h"
#include "logger.h"
#include "mesh.h"
#include "meshgen.h"
#include "rng.h"


// default PineTreeParams values
#define DEF_RAD_SCALE		0.025
#define DEF_SCALE			0.32
#define DEF_SCALE_OFFS		0.05
#define DEF_ANGLE			60.0
#define DEF_MIN_Y			0.15
#define DEF_MAX_Y			0.96
#define DEF_SPIRAL_TURNS	11.0
#define DEF_DISTR_BIAS		0.9
// light parameters
#define DEF_BLINK_SPEED		2.0
#define DEF_LT_DISTR_BIAS	2.0
#define DEF_LT_TURNS		8.0
#define DEF_LT_SPAWN_MIN	(DEF_MIN_Y + 0.15)
#define DEF_LT_SPAWN_MAX	(DEF_MAX_Y - 0.05)
// needle parameters
#define DEF_NEEDLES_DENSITY	2.5

static Object *gen_tree_rec(PineTree *xtree, Object *par, PineTreeBranch *par_br,
		float par_spawn_pos, const Vector3 &pos, const Quaternion &rot, float size,
		int *nbranches, int level, int max_iter, int conesub);


void PineTree::init_gen()
{
	gen_params.scale = DEF_SCALE;
	gen_params.rad_scale = DEF_RAD_SCALE;
	gen_params.scale_offs = DEF_SCALE_OFFS;
	gen_params.angle = DEF_ANGLE;
	gen_params.spawn_min = DEF_MIN_Y;
	gen_params.spawn_max = DEF_MAX_Y;
	gen_params.spiral_turns = DEF_SPIRAL_TURNS;
	gen_params.distr_bias = DEF_DISTR_BIAS;

	gen_params.lt_blink_speed = DEF_BLINK_SPEED;
	gen_params.lt_distr_bias = DEF_LT_DISTR_BIAS;
	gen_params.lt_turns = DEF_LT_TURNS;
	gen_params.lt_spawn_min = DEF_LT_SPAWN_MIN;
	gen_params.lt_spawn_max = DEF_LT_SPAWN_MAX;

	gen_params.needles_density = DEF_NEEDLES_DENSITY;
}

bool PineTree::generate(float size, ...)
{
	va_list ap;

	if(!init()) {
		return false;
	}

	// clear everything first
	clear();
	needles_mesh.clear();
	this->size = size;

	std::vector<int> nbranches;
	va_start(ap, size);
	int n;
	do {
		n = va_arg(ap, int);
		nbranches.push_back(n);
	} while(n);
	va_end(ap);

	int num_levels = (int)nbranches.size();

	Vector3 pos;
	Quaternion rot;

	Object *obj = gen_tree_rec(this, 0, 0, 0, pos, rot, size, &nbranches[0], 0, num_levels, 8);
	if(!obj) {
		return false;
	}
	root = obj;
	root_branch = branches.empty() ? 0 : branches.front();
	if(root_branch) {
		root_branch->calc_bounds();
	}
	info_log("generated tree with %ld polygons\n", root->count_polys());

	// now let's apply all transformations to the individual meshes ...
	root->apply_xform();

	// ... and then merge them into one large mesh
	std::list<Mesh*> mlist = root->get_all_meshes();
	std::list<Mesh*>::iterator it = mlist.begin();

	Mesh *ubermesh = root->get_mesh();
	while(it != mlist.end()) {
		Mesh *mesh = *it++;
		if(mesh == ubermesh) {
			continue;
		}

		ubermesh->append(*mesh);
	}

	root = new Object;
	root->set_mesh(ubermesh);
	root->material.tex[0] = bark_tex;
	root->material.tex_xform.set_scaling(Vector3(1.0, 6.0, 1.0));

	// go through the scene meshes, find the ubermesh and remove it to save it from destruction
	for(size_t i=0; i<meshes.size(); i++) {
		if(meshes[i] == ubermesh) {
			meshes.erase(meshes.begin() + i);
			break;
		}
	}

	Scene::clear();	// clear all the previous objects / meshes
	objects.push_back(root);
	meshes.push_back(ubermesh);


	// calculate approximate ambient occlusion for all branches
	for(size_t i=0; i<branches.size(); i++) {
		PineTreeBranch *br = branches[i];

		for(int j=0; j<2; j++) {
			float tree_rad = radius(br->end[j].y);
			float dist = tree_rad - Vector3(br->end[j].x, 0, br->end[j].z).length();

			float ao = 1.3 - dist;
			br->amb_occ[j] = ao > 1.0 ? 1.0 : (ao < 0.0 ? 0.0 : ao);
		}
	}


	// generate the needles mesh (2 cones back&front each)
	gen_cone(&needles_mesh, 0.8, 1.0, 5, 0);
	Mesh inner_cone;
	gen_cone(&inner_cone, 0.3, 1.0, 4, 0);
	needles_mesh.append(inner_cone);

	Mesh backfaces = needles_mesh;
	backfaces.flip();
	needles_mesh.append(backfaces);

	Matrix4x4 xform;
	xform.set_rotation(Vector3(0, 0, M_PI));
	xform.translate(Vector3(0, -1, 0));
	needles_mesh.apply_xform(xform);

	// add color attributes to the needle mesh vertices which we'll use for
	// ambient occlusion and needle position along the branch
	int needle_nverts = needles_mesh.get_vertex_count();
	needles_mesh.set_attrib_data(MESH_ATTR_COLOR, 4, needle_nverts);

	// XXX optimization experiment -----------------
	// generate the branch needle meshes by concatenating a bunch of needle_meshes created above
	int nleafbr = leaf_branches.size();
	needles_lbr_meshes = (Mesh*)new char[nleafbr * sizeof *needles_lbr_meshes];

	for(int i=0; i<nleafbr; i++) {
		PineTreeBranch *br = leaf_branches[i];

		int num_needles = ceil(br->size * gen_params.needles_density);

		for(int j=0; j<num_needles; j++) {
			float t = lerp(gen_params.spawn_min, gen_params.spawn_max, (float)j / (float)num_needles);

			float size = branch_size(&gen_params, br->size, t);
			float ao = lerp(br->amb_occ[0], br->amb_occ[1], t);

			Matrix4x4 matrix = br->xform;
			matrix.translate(Vector3(0, t * br->size, 0));
			matrix.scale(Vector3(size, size, size));

			Mesh mesh = needles_mesh;
			mesh.apply_xform(matrix);
			/* Use the vertex color attribute to pass per-needle information:
			 * - red = normalized position along the branch [0, 1]
			 * - alpha = ambient occlusion
			 */
			Vector4 *vcol = (Vector4*)mesh.get_attrib_data(MESH_ATTR_COLOR);
			for(int k=0; k<needle_nverts; k++) {
				vcol[k] = Vector4(t, 0, 0, ao);
			}

			if(j == 0) {
				new (needles_lbr_meshes + i) Mesh(mesh);
			} else {
				needles_lbr_meshes[i].append(mesh);
			}
		}
	}
	return true;
}


bool PineTree::generate_lights(int num, float turns)
{
	static const Vector3 lcol = Vector3(1.0, 0.8, 0.5);

	float y0 = gen_params.lt_spawn_min;
	float y1 = gen_params.lt_spawn_max;

	for(int i=0; i<num; i++) {
		float t = (float)i / (float)num;
		float yjitter = rng_frand() * 0.1 - 0.05;
		float y = (lerp(y0, y1, pow(t, gen_params.lt_distr_bias)) + yjitter) * size;
		float angle = 2.0 * M_PI * t * turns;

		float rad = std::max(radius(y) - 0.25, 0.01);
		Vector3 pos = Vector3(cos(angle) * rad, y, sin(angle) * rad);

		TreeLight lt;
		lt.alive = true;
		lt.pos = pos;
		lt.intensity = 1.0;
		lt.color = lcol;
		lt.blink_offs = 2.0 * M_PI * rng_frand();
		lt.blink_speed = gen_params.lt_blink_speed;
		lt.size = 0.2;
		lights.push_back(lt);
	}
	return true;
}


/* RECURSIVE TREE GENERATION
 * this function pretty much does everything
 */
static Object *gen_tree_rec(PineTree *xtree, Object *par, PineTreeBranch *par_br,
		float par_spawn_pos, const Vector3 &pos, const Quaternion &rot, float size,
		int *nbranches, int level, int max_iter, int conesub)
{
	if(level >= max_iter || size < 1e-3) {
		return 0;
	}
	if(conesub < 3) conesub = 3;

	PineTreeParams *param = &xtree->gen_params;

	// create the mesh
	Mesh *mesh = new Mesh;
	gen_cone(mesh, param->rad_scale * size, size, conesub, 1);
	xtree->meshes.push_back(mesh);

	// create the scene node
	Object *obj = new Object;
	if(par) {
		par->add_child(obj);
	}
	obj->set_mesh(mesh);
	obj->set_position(pos);
	obj->set_rotation(rot);
	xtree->objects.push_back(obj);

	int num_branches = nbranches[level];


	// create the branch data structure
	PineTreeBranch *br = new PineTreeBranch;
	br->tree = xtree;
	br->parent = par_br;
	if(par_br) {
		par_br->children.push_back(br);
	}
	br->spawn_pt = par_spawn_pos;
	xtree->branches.push_back(br);

	Matrix4x4 xform;
	obj->get_xform(0, &xform);
	br->xform = xform;

	br->end[0] = Vector3(0, 0, 0).transformed(xform);
	br->end[1] = Vector3(0, size, 0).transformed(xform);
	br->size = (br->end[1] - br->end[0]).length();

	//if(num_branches) {
	if(br->parent) {
		// it's a leaf branch used to spawn pine needles
		xtree->leaf_branches.push_back(br);
	} else {
		xtree->inner_branches.push_back(br);
	}
	br->calc_bounds();

	// recursively create all the sub-branches
	float spawn_range = param->spawn_max - param->spawn_min;
	float jitter_mag = 0.5 * spawn_range / param->spiral_turns;

	int child_conesub = conesub * 2 / 3;

	for(int i=0; i<num_branches; i++) {

		// t is the place of the sprout along the spiral circling the parent branch
		float t = (float)i / (float)num_branches;
		t = pow(t, param->distr_bias);	// bias towards either end (to bottom for bias > 1)
		// calculate the actual displacement along the parent branch
		float y = param->spawn_min + spawn_range * t;
		// and jitter it to stop looking like a spiral
		y += (rng_frand() * 2.0 - 1.0) * jitter_mag;

		float br_pos = y * size;
		float br_size = branch_size(param, size, t);
		float br_angle = fmod(t * param->spiral_turns, 1.0) * 2.0 * M_PI;

		// determine where to place it and under which orientation
		Vector3 pos = Vector3(0, br_pos, 0);
		Quaternion q;
		q.rotate(Vector3(0, 1, 0), br_angle);
		q.rotate(Vector3(1, 0, 0), DEG_TO_RAD(param->angle));

		gen_tree_rec(xtree, obj, br, y, pos, q, br_size, nbranches, level + 1, max_iter, child_conesub);
	}

	return obj;
}
